"""TcEx Framework Module"""

# standard library
import inspect
import platform
import signal
import threading
from dataclasses import dataclass
from pathlib import Path

# first-party
from tcex.api.api import API
from tcex.app.app import App
from tcex.exit.exit import Exit, ExitCode
from tcex.input.input import Input
from tcex.logger.logger import Logger
from tcex.logger.trace_logger import TraceLogger
from tcex.pleb.cached_property import cached_property
from tcex.pleb.proxies import proxies
from tcex.pleb.scoped_property import scoped_property
from tcex.registry import registry
from tcex.requests_external import ExternalSession, RequestsExternal
from tcex.requests_tc import RequestsTc, TcSession
from tcex.util import Util


@dataclass
class _Session:
    """Session dataclass."""

    external: ExternalSession
    tc: TcSession


class TcEx:
    """Provides functionality for all types of TxEx Apps.

    Args:
        config (dict, kwargs): A dictionary containing configuration items typically used by
            external Apps.
        config_file (str, kwargs): A filename containing JSON configuration items typically used
            by external Apps.
    """

    def __init__(self, **kwargs):
        """Initialize instance properties."""
        # catch interrupt signals specifically based on thread name
        signal.signal(signal.SIGINT, self._signal_handler)
        if platform.system() != 'Windows':
            signal.signal(signal.SIGHUP, self._signal_handler)
        signal.signal(signal.SIGTERM, self._signal_handler)

        # properties
        self._log: TraceLogger  # allow logger to be overridden

        # initialize TcEx/App inputs
        self.inputs = Input(kwargs.get('config') or {}, kwargs.get('config_file'))

        # add methods to registry
        registry.register(self)
        registry.add_service(Input, self.inputs)
        registry.add_service(App, self.app)
        registry.add_service(Logger, self.logger)
        registry.add_service(RequestsTc, self.requests_tc)

        # log standard App info early so it shows at the top of the logfile
        self.logger.log_info(self.inputs.model_tc)

    def _signal_handler(self, signal_interrupt: int, _):
        """Handle signal interrupt."""
        call_file = Path(inspect.stack()[1][0].f_code.co_filename).name
        call_module = inspect.stack()[1][0].f_globals['__name__'].replace('Functions.', '', 1)
        call_line = inspect.stack()[1][0].f_lineno
        self.log.error(
            f'App interrupted - file: {call_file}, method: {call_module}, line: {call_line}.'
        )
        exit_code = ExitCode.SUCCESS
        if threading.current_thread().name == 'MainThread' and signal_interrupt in (2, 15):
            exit_code = ExitCode.FAILURE

        self.exit.exit(exit_code, 'The App received an interrupt signal and will now exit.')

    @cached_property
    def api(self) -> API:
        """Return instance of API."""
        return API(self.inputs, self.session.tc)

    @cached_property
    def app(self) -> App:
        """Return instance of App."""
        return App(self.inputs.module_app_model, self.proxies)

    @registry.factory(Exit)
    @scoped_property
    def exit(self) -> Exit:
        """Return an instance of Exit."""
        return Exit(self.inputs)

    @property
    def log(self) -> TraceLogger:
        """Return a valid logger."""
        if self._log is None:
            self._log = self.logger.log
        return self._log

    @log.setter
    def log(self, log: TraceLogger):
        """Return a valid logger."""
        if isinstance(log, TraceLogger):
            self._log = log

    @cached_property
    def logger(self) -> Logger:
        """Return logger."""
        _logger = Logger(logger_name='tcex')

        # set logger to prevent recursion issue on get_session
        self._log = _logger.log

        # add api handler
        if self.inputs.contents.get('tc_token') is not None and self.inputs.contents.get(
            'tc_log_to_api'
        ) in (True, 'true'):
            _logger.add_api_handler(
                session_tc=self.requests_tc.get_session(),
                level=self.inputs.model_tc.tc_log_level,
            )

        # add rotating log handler
        _logger.add_rotating_file_handler(
            name='rfh',
            filename=self.inputs.model_tc.tc_log_file,
            path=self.inputs.model_tc.tc_log_path,
            backup_count=self.inputs.model_tc.tc_log_backup_count,
            max_bytes=self.inputs.model_tc.tc_log_max_bytes,
            level=self.inputs.model_tc.tc_log_level,
        )

        # set logging level
        _logger.update_handler_level(level=self.inputs.model_tc.tc_log_level)
        _logger.log.setLevel(_logger.log_level(self.inputs.model_tc.tc_log_level))

        # replay cached log events
        _logger.replay_cached_events(handler_name='cache')

        return _logger

    @cached_property
    def proxies(self) -> dict:
        """Format the proxy configuration for Python Requests module.

        Generates a dictionary for use with the Python Requests module format
        when proxy is required for remote connections.

        **Example Response**
        ::

            {'http': 'http://user:pass@10.10.1.10:3128/'}
        """
        return proxies(
            proxy_host=self.inputs.model_tc.tc_proxy_host,
            proxy_port=self.inputs.model_tc.tc_proxy_port,
            proxy_user=self.inputs.model_tc.tc_proxy_username,
            proxy_pass=self.inputs.model_tc.tc_proxy_password,
        )

    @cached_property
    def requests_external(self) -> RequestsExternal:
        """Return instance of RequestsSession."""
        return RequestsExternal(self.inputs.module_requests_session_model, self.proxies)

    @cached_property
    def requests_tc(self) -> RequestsTc:
        """Return instance of RequestsSession."""
        return RequestsTc(self.inputs.module_requests_session_model)

    @cached_property
    def session(self) -> _Session:
        """Return Session object.

        Add interface to access Sessions:

        * self.tcex.session.external
        * self.tcex.session.tc
        """
        return _Session(external=self.requests_external.session, tc=self.requests_tc.session)

    @cached_property
    def util(self) -> Util:
        """Return instance of Util."""
        return Util()



# standard library
import os
from abc import ABC

# third-party
from requests import Session

# first-party
from tcex.input.field_type.sensitive import Sensitive
from tcex.requests_tc.auth.hmac_auth import HmacAuth


class OptionsABC(ABC):  # noqa: B024
    """Abstract base class for all Options classes."""

    def __init__(self, **kwargs):
        """Initialize the OptionsABC class."""
        self._options = kwargs

    @property
    def session(self) -> Session:
        """Return Session configured for TC API."""
        _session = Session()
        _session.auth = HmacAuth(
            os.getenv('TC_API_ACCESS_ID', ''), Sensitive(os.getenv('TC_API_SECRET_KEY', ''))
        )
        return _session



"""TcEx Framework Module"""

# standard library
from collections.abc import Generator, Iterator
from typing import TYPE_CHECKING

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.groups.group_model import GroupModel
from tcex.api.tc.v3.object_abc import ObjectABC
from tcex.api.tc.v3.object_collection_abc import ObjectCollectionABC
from tcex.api.tc.v3.victim_assets.victim_asset_filter import VictimAssetFilter
from tcex.api.tc.v3.victim_assets.victim_asset_model import VictimAssetModel, VictimAssetsModel

if TYPE_CHECKING:  # pragma: no cover
    # first-party
    from tcex.api.tc.v3.groups.group import Group  # CIRCULAR-IMPORT


class VictimAsset(ObjectABC):
    """VictimAssets Object.

    Args:
        account_name (str, kwargs): The network name.
        address (str, kwargs): The email address associated with the E-Mail Address asset.
        address_type (str, kwargs): The type of the E-Mail Address asset.
        associated_groups (Groups, kwargs): A list of groups that this victim asset is associated
            with.
        network_type (str, kwargs): The type of network.
        phone (str, kwargs): The phone number of the asset.
        social_network (str, kwargs): The type of social network.
        type (str, kwargs): Type of victim asset.
        victim_id (int, kwargs): Victim id of victim asset.
        website (str, kwargs): The website of the asset.
    """

    def __init__(self, **kwargs):
        """Initialize instance properties."""
        super().__init__(kwargs.pop('session', None))

        # properties
        self._model: VictimAssetModel = VictimAssetModel(**kwargs)
        self._nested_field_name = 'victimAssets'
        self._nested_filter = 'has_victim_asset'
        self.type_ = 'Victim Asset'

    @property
    def _api_endpoint(self) -> str:
        """Return the type specific API endpoint."""
        return ApiEndpoints.VICTIM_ASSETS.value

    @property
    def model(self) -> VictimAssetModel:
        """Return the model data."""
        return self._model

    @model.setter
    def model(self, data: dict | VictimAssetModel):
        """Create model using the provided data."""
        if isinstance(data, type(self.model)):
            # provided data is already a model, nothing required to change
            self._model = data
        elif isinstance(data, dict):
            # provided data is raw response, load the model
            self._model = type(self.model)(**data)
        else:
            ex_msg = f'Invalid data type: {type(data)} provided.'
            raise RuntimeError(ex_msg)  # noqa: TRY004

    @property
    def as_entity(self) -> dict:
        """Return the entity representation of the object."""
        value = []

        if self.model.type is not None:
            if self.model.type.lower() == 'phone':
                if self.model.phone:
                    value.append(self.model.phone)
            elif self.model.type.lower() == 'socialnetwork':
                if self.model.social_network:
                    value.append(self.model.social_network)
                if self.model.account_name:
                    value.append(self.model.account_name)
            elif self.model.type.lower() == 'networkaccount':
                if self.model.network_type:
                    value.append(self.model.network_type)
                if self.model.account_name:
                    value.append(self.model.account_name)
            elif self.model.type.lower() == 'emailaddress':
                if self.model.address_type:
                    value.append(self.model.address_type)
                if self.model.address:
                    value.append(self.model.address)
            elif self.model.type.lower() == 'website' and self.model.website:
                value.append(self.model.website)

        value = ' : '.join(value) if value else ''
        type_ = f'Victim Asset : {self.model.type}'

        return {'type': type_, 'id': self.model.id, 'value': value}

    @property
    def associated_groups(self) -> Generator['Group', None, None]:
        """Yield Group from Groups."""
        # first-party
        from tcex.api.tc.v3.groups.group import Groups

        yield from self._iterate_over_sublist(Groups)  # type: ignore

    def stage_associated_group(self, data: dict | ObjectABC | GroupModel):
        """Stage group on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = GroupModel(**data)

        if not isinstance(data, GroupModel):
            ex_msg = 'Invalid type passed in to stage_associated_group'
            raise RuntimeError(ex_msg)  # noqa: TRY004
        data._staged = True  # noqa: SLF001
        self.model.associated_groups.data.append(data)  # type: ignore


class VictimAssets(ObjectCollectionABC):
    """VictimAssets Collection.

    # Example of params input
    {
        "result_limit": 100,  # Limit the retrieved results.
        "result_start": 10,  # Starting count used for pagination.
        "fields": ["caseId", "summary"]  # Select additional return fields.
    }

    Args:
        session (Session): Session object configured with TC API Auth.
        tql_filters (list): List of TQL filters.
        params (dict): Additional query params (see example above).
    """

    def __init__(self, **kwargs):
        """Initialize instance properties."""
        super().__init__(
            kwargs.pop('session', None), kwargs.pop('tql_filter', None), kwargs.pop('params', None)
        )
        self._model = VictimAssetsModel(**kwargs)
        self.type_ = 'victim_assets'

    def __iter__(self) -> Iterator[VictimAsset]:
        """Return CM objects."""
        return self.iterate(base_class=VictimAsset)  # type: ignore

    @property
    def _api_endpoint(self) -> str:
        """Return the type specific API endpoint."""
        return ApiEndpoints.VICTIM_ASSETS.value

    @property
    def filter(self) -> VictimAssetFilter:
        """Return the type specific filter object."""
        return VictimAssetFilter(self.tql)



"""TcEx Framework Module"""

# third-party
from requests import Session

# first-party
from tcex.api.tc.v3.intel_requirements.categories.category import Categories, Category
from tcex.api.tc.v3.intel_requirements.intel_requirement import IntelRequirement, IntelRequirements
from tcex.api.tc.v3.intel_requirements.results.result import Result, Results
from tcex.api.tc.v3.intel_requirements.subtypes.subtype import Subtype, Subtypes


class IR:
    """Intel Requirement

    Args:
        session: An configured instance of request.Session with TC API Auth.
    """

    def __init__(self, session: Session):
        """Initialize instance properties."""
        self.session = session

    def categories(self, **kwargs) -> Categories:
        """Return a instance of Categories object."""
        return Categories(session=self.session, **kwargs)

    def category(self, **kwargs) -> Category:
        """Return a instance of Category object."""
        return Category(session=self.session, **kwargs)

    def intel_requirement(self, **kwargs) -> IntelRequirement:
        """Return a instance of Intel Requirement object."""
        return IntelRequirement(session=self.session, **kwargs)

    def intel_requirements(self, **kwargs) -> IntelRequirements:
        """Return a instance of Intel Requirements object."""
        return IntelRequirements(session=self.session, **kwargs)

    def result(self, **kwargs) -> Result:
        """Return a instance of Result object."""
        return Result(session=self.session, **kwargs)

    def results(self, **kwargs) -> Results:
        """Return a instance of Results object."""
        return Results(session=self.session, **kwargs)

    def subtype(self, **kwargs) -> Subtype:
        """Return a instance of Subtype object."""
        return Subtype(session=self.session, **kwargs)

    def subtypes(self, **kwargs) -> Subtypes:
        """Return a instance of Subtypes object."""
        return Subtypes(session=self.session, **kwargs)



"""TcEx Framework Module"""

# standard library
from collections.abc import Generator, Iterator
from typing import TYPE_CHECKING

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.artifacts.artifact_model import ArtifactModel
from tcex.api.tc.v3.notes.note_model import NoteModel
from tcex.api.tc.v3.object_abc import ObjectABC
from tcex.api.tc.v3.object_collection_abc import ObjectCollectionABC
from tcex.api.tc.v3.security.assignee_model import AssigneeModel
from tcex.api.tc.v3.security.assignee_user_group_model import AssigneeUserGroupModel
from tcex.api.tc.v3.security.assignee_user_model import AssigneeUserModel
from tcex.api.tc.v3.security.user_groups.user_group_model import UserGroupModel
from tcex.api.tc.v3.security.users.user_model import UserModel
from tcex.api.tc.v3.tasks.task_filter import TaskFilter
from tcex.api.tc.v3.tasks.task_model import TaskModel, TasksModel

if TYPE_CHECKING:  # pragma: no cover
    # first-party
    from tcex.api.tc.v3.artifacts.artifact import Artifact  # CIRCULAR-IMPORT
    from tcex.api.tc.v3.notes.note import Note  # CIRCULAR-IMPORT


class Task(ObjectABC):
    """Tasks Object.

    Args:
        artifacts (Artifacts, kwargs): A list of Artifacts corresponding to the Task.
        assignee (Assignee, kwargs): The user or group Assignee object for the Task.
        case_id (int, kwargs): The **case id** for the Task.
        case_xid (str, kwargs): The **case xid** for the Task.
        completed_date (str, kwargs): The completion date of the Task.
        dependent_on_id (int, kwargs): The ID of another Task that this Task is dependent upon.
        description (str, kwargs): The **description** for the Task.
        due_date (str, kwargs): The due date of the Task.
        duration_type (str, kwargs): The **duration type** for the Task.
        name (str, kwargs): The **name** for the Task.
        notes (Notes, kwargs): A list of Notes corresponding to the Task.
        required (bool, kwargs): Flag indicating whether or not the task is required.
        status (str, kwargs): The **status** for the Task.
        workflow_phase (int, kwargs): The phase of the workflow.
        workflow_step (int, kwargs): The step of the workflow.
        xid (str, kwargs): The **xid** for the Task.
    """

    def __init__(self, **kwargs):
        """Initialize instance properties."""
        super().__init__(kwargs.pop('session', None))

        # properties
        self._model: TaskModel = TaskModel(**kwargs)
        self._nested_field_name = 'tasks'
        self._nested_filter = 'has_task'
        self.type_ = 'Task'

    @property
    def _api_endpoint(self) -> str:
        """Return the type specific API endpoint."""
        return ApiEndpoints.TASKS.value

    @property
    def model(self) -> TaskModel:
        """Return the model data."""
        return self._model

    @model.setter
    def model(self, data: dict | TaskModel):
        """Create model using the provided data."""
        if isinstance(data, type(self.model)):
            # provided data is already a model, nothing required to change
            self._model = data
        elif isinstance(data, dict):
            # provided data is raw response, load the model
            self._model = type(self.model)(**data)
        else:
            ex_msg = f'Invalid data type: {type(data)} provided.'
            raise RuntimeError(ex_msg)  # noqa: TRY004

    @property
    def as_entity(self) -> dict:
        """Return the entity representation of the object."""
        type_ = self.type_

        return {'type': type_, 'id': self.model.id, 'value': self.model.name}

    @property
    def artifacts(self) -> Generator['Artifact', None, None]:
        """Yield Artifact from Artifacts."""
        # first-party
        from tcex.api.tc.v3.artifacts.artifact import Artifacts

        yield from self._iterate_over_sublist(Artifacts)  # type: ignore

    @property
    def notes(self) -> Generator['Note', None, None]:
        """Yield Note from Notes."""
        # first-party
        from tcex.api.tc.v3.notes.note import Notes

        yield from self._iterate_over_sublist(Notes)  # type: ignore

    def stage_artifact(self, data: dict | ObjectABC | ArtifactModel):
        """Stage artifact on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = ArtifactModel(**data)

        if not isinstance(data, ArtifactModel):
            ex_msg = 'Invalid type passed in to stage_artifact'
            raise RuntimeError(ex_msg)  # noqa: TRY004
        data._staged = True  # noqa: SLF001
        self.model.artifacts.data.append(data)  # type: ignore

    def stage_assignee(
        self,
        type: str,  # noqa: A002
        data: dict | ObjectABC | AssigneeModel | UserModel | UserGroupModel,
    ):
        """Stage artifact on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif type.lower() == 'user' and isinstance(data, dict):
            data = AssigneeModel(type='User', data=data)  # type: ignore
        elif type.lower() == 'group' and isinstance(data, dict):
            data = AssigneeModel(type='Group', data=data)  # type: ignore

        if not isinstance(
            data,
            AssigneeModel | AssigneeUserModel | AssigneeUserGroupModel | UserModel | UserGroupModel,
        ):
            ex_msg = 'Invalid type passed in to stage_assignee'
            raise RuntimeError(ex_msg)  # noqa: TRY004

        if isinstance(data, AssigneeModel):
            self.model.assignee = data
        elif isinstance(
            data, AssigneeUserModel | AssigneeUserGroupModel | UserModel | UserGroupModel
        ):
            self.model.assignee.data = data  # type: ignore
        self.model.assignee.data._staged = True  # type: ignore  # noqa: SLF001

    def stage_note(self, data: dict | ObjectABC | NoteModel):
        """Stage note on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = NoteModel(**data)

        if not isinstance(data, NoteModel):
            ex_msg = 'Invalid type passed in to stage_note'
            raise RuntimeError(ex_msg)  # noqa: TRY004
        data._staged = True  # noqa: SLF001
        self.model.notes.data.append(data)  # type: ignore


class Tasks(ObjectCollectionABC):
    """Tasks Collection.

    # Example of params input
    {
        "result_limit": 100,  # Limit the retrieved results.
        "result_start": 10,  # Starting count used for pagination.
        "fields": ["caseId", "summary"]  # Select additional return fields.
    }

    Args:
        session (Session): Session object configured with TC API Auth.
        tql_filters (list): List of TQL filters.
        params (dict): Additional query params (see example above).
    """

    def __init__(self, **kwargs):
        """Initialize instance properties."""
        super().__init__(
            kwargs.pop('session', None), kwargs.pop('tql_filter', None), kwargs.pop('params', None)
        )
        self._model = TasksModel(**kwargs)
        self.type_ = 'tasks'

    def __iter__(self) -> Iterator[Task]:
        """Return CM objects."""
        return self.iterate(base_class=Task)  # type: ignore

    @property
    def _api_endpoint(self) -> str:
        """Return the type specific API endpoint."""
        return ApiEndpoints.TASKS.value

    @property
    def filter(self) -> TaskFilter:
        """Return the type specific filter object."""
        return TaskFilter(self.tql)



"""TcEx Framework Module"""

# standard library
from typing import Literal

# third-party
from rich import print as print_
from rich.panel import Panel
from rich.table import Table

AlignMethod = Literal['left', 'center', 'right']


class RenderTable:
    """Render Table"""

    title_align: AlignMethod = 'left'

    @classmethod
    def key_value(
        cls,
        title: str,
        kv_data: dict[str, str | None] | list[dict[str, str | None]],
        border_style: str = '',
        key_style: str = 'dodger_blue1',
        key_width: int = 20,
        value_style: str = 'bold',
        value_width: int = 80,
    ):
        """Render key/value table.

        Accepts the following structuresL
        {
            'key': 'value',
        }
        or
        [
            {
                'key': 'my key',
                'value': 'my_value'
            }
        ]
        """
        table = Table(
            border_style=border_style,
            expand=True,
            show_edge=False,
            show_header=False,
        )

        table.add_column(
            'key',
            justify='left',
            max_width=key_width,
            min_width=key_width,
            style=key_style,
        )
        table.add_column(
            'value',
            justify='left',
            max_width=value_width,
            min_width=value_width,
            style=value_style,
        )

        if isinstance(kv_data, dict):
            for key, value in kv_data.items():
                table.add_row(key, value)
        elif isinstance(kv_data, list):
            for item in kv_data:
                table.add_row(item['key'], item['value'])

        # render panel->table
        if kv_data:
            print_(
                Panel(table, border_style=border_style, title=title, title_align=cls.title_align)
            )



"""TcEx Framework Module"""

# third-party
from pydantic import BaseModel, Extra, Field, validator

# first-party
from tcex.pleb.cached_property import cached_property
from tcex.util import Util
from tcex.util.string_operation import CamelString


class ExtraModel(
    BaseModel,
    extra=Extra.forbid,
    title='Extra Model',
    validate_assignment=True,
):
    """Model Definition"""

    comment: str = Field('', description='The comment for the extra.')
    tql_type: str = Field(..., description='The TQL type.')
    typing_type: str = Field(..., description='The Python typing hint type.')


class FilterModel(
    BaseModel,
    alias_generator=Util().snake_to_camel,
    arbitrary_types_allowed=True,
    extra=Extra.forbid,
    keep_untouched=(cached_property,),
    title='Property Model',
    validate_assignment=True,
):
    """Model Definition"""

    description: str = Field('No description provided.', description='Description of the filter.')
    groupable: bool = Field(default=False, description='True if the filter is able to be grouped.')
    keyword: CamelString = Field(..., description='The filter keyword.')
    name: CamelString = Field(..., description='The name of the filter.')
    targetable: bool = Field(
        default=False, description='True if the filter is able to be targeted.'
    )
    type: str = Field(..., description='The type of the filter.')
    units: dict | None = Field(
        default=None,
        description='The units of the filter. This is only used for filters that are numeric.',
    )

    @validator('keyword', 'name', always=True, pre=True)
    @classmethod
    def _camel_string(cls, v):
        """Convert to CamelString."""
        return CamelString(v)

    @validator('description', always=True, pre=True)
    @classmethod
    def _fix_first_letter_of_sentence(cls, v):
        """Convert to CamelString."""
        if v:
            # ensure the first letter of the sentence is capitalized
            v = v[0].capitalize() + v[1:]
        return v

    @cached_property
    def extra(self) -> ExtraModel:
        """Return the extra model."""
        extra_data = self.__extra_data()
        # print(self.name, self.type, extra_data)
        return ExtraModel(**extra_data)  # type: ignore

    def __extra_data(self) -> dict:
        """Return the extra data."""
        typing_type_data = self.__calculate_typing_type(self.type)
        typing_type = typing_type_data['typing_type']
        tql_type = self.__calculate_tql_type(typing_type_data['base_type'])
        return {
            'comment': self.__calculate_comment(self.keyword),
            'typing_type': typing_type,
            'tql_type': tql_type,
        }

    @classmethod
    def __calculate_comment(cls, keyword: str) -> str:
        """Return the calculated filter type."""
        if keyword in ['id', 'type']:
            return '  # noqa: A002'
        return ''

    @classmethod
    def __calculate_typing_type(cls, type_: str) -> dict[str, str]:
        """Return the calculated filter type."""
        # hint types for the filter method
        filter_type_map = {
            'Assignee': {'base_type': 'str', 'typing_type': 'list | str'},
            'BigInteger': {'base_type': 'int', 'typing_type': 'int | list'},
            'Boolean': {'base_type': 'bool', 'typing_type': 'bool'},
            'Date': {'base_type': 'str', 'typing_type': 'Arrow | datetime | int | str'},
            'DateTime': {'base_type': 'str', 'typing_type': 'Arrow | datetime | int | str'},
            'Double': {'base_type': 'float', 'typing_type': 'float | list'},
            'Enum': {'base_type': 'str', 'typing_type': 'list | str'},
            'EnumToInteger': {'base_type': 'str', 'typing_type': 'list | str'},
            'Integer': {'base_type': 'int', 'typing_type': 'int | list'},
            'Long': {'base_type': 'int', 'typing_type': 'int | list'},
            'String': {'base_type': 'str', 'typing_type': 'list | str'},
            'StringCIDR': {'base_type': 'str', 'typing_type': 'list | str'},
            'StringLower': {'base_type': 'str', 'typing_type': 'list | str'},
            'StringUpper': {'base_type': 'str', 'typing_type': 'list | str'},
            'Undefined': {'base_type': 'str', 'typing_type': 'list | str'},
            'User': {'base_type': 'str', 'typing_type': 'list | str'},
        }
        hint_type = filter_type_map.get(type_)

        # fail on any unknown types (core added something new)
        if hint_type is None:
            ex_msg = f'Invalid type of {type_} (Core team added something new?)'
            raise RuntimeError(ex_msg)

        return hint_type

    @classmethod
    def __calculate_tql_type(cls, typing_type: str):
        """Return the calculated tql type."""
        # tql types for the add_filter method call
        tql_type_map = {
            'bool': 'TqlType.BOOLEAN',
            'float': 'TqlType.FLOAT',
            'int': 'TqlType.INTEGER',
            'str': 'TqlType.STRING',
        }
        return tql_type_map.get(typing_type)



"""TcEx Framework Module"""

# third-party
from pydantic import BaseModel, Extra, Field, PrivateAttr

# first-party
from tcex.api.tc.v3.v3_model_abc import V3ModelABC
from tcex.util import Util


class FileActionModel(
    V3ModelABC,
    alias_generator=Util().snake_to_camel,
    extra=Extra.allow,
    title='File Action Model',
    validate_assignment=True,
):
    """File Action Model"""

    relationship: str = Field(
        ...,
        description='The File Action type.',
        methods=['POST', 'PUT'],
        title='relationship',
        read_only=False,
    )
    indicator: 'IndicatorModel' = Field(
        ...,
        description='The **indicator** related to the FileAction.',
        methods=['POST', 'PUT'],
        title='indicator',
        read_only=False,
    )


class FileActionsModel(
    BaseModel,
    title='File Actions Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """File Actions Model"""

    _mode_support = PrivateAttr(default=True)

    count: int | None = Field(None, description='The number of file actions.')

    data: list[FileActionModel] | None = Field(
        [],
        description='The data for the File Actions.',
        methods=['POST', 'PUT'],
        title='data',
    )

    mode: str = Field(
        'append',
        description='The PUT mode for nested objects (append, delete, replace). Default: append',
        methods=['POST', 'PUT'],
        title='append',
    )


# first-party
from tcex.api.tc.v3.indicators.indicator import IndicatorModel

FileActionModel.update_forward_refs()
FileActionsModel.update_forward_refs()



"""TcEx Framework Module"""

# standard library
import contextlib
import logging

# third-party
import urllib3
from requests import Response, Session, adapters
from urllib3.util.retry import Retry

from ..util.requests_to_curl import RequestsToCurl  # type: ignore
from ..util.util import Util  # type: ignore
from .auth.hmac_auth import HmacAuth
from .auth.tc_auth import TcAuth
from .auth.token_auth import TokenAuth

# get logger
_logger = logging.getLogger(__name__.split('.', maxsplit=1)[0])

# disable ssl warning message
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)  # type: ignore


class TcSession(Session):
    """ThreatConnect REST API Requests Session"""

    def __init__(
        self,
        auth: HmacAuth | TokenAuth | TcAuth,
        base_url: str | None = None,
        log_curl: bool | None = False,
        proxies: dict[str, str] | None = None,
        proxies_enabled: bool | None = False,
        user_agent: dict[str, str] | None = None,
        verify: bool | str | None = True,
    ):
        """Initialize the Class properties."""
        super().__init__()
        self.base_url = base_url.strip('/') if base_url is not None else base_url
        self.log = _logger
        self.log_curl = log_curl

        # properties
        self.requests_to_curl = RequestsToCurl()
        self.util = Util()

        # configure auth
        self.auth = auth

        # configure optional headers
        if user_agent:
            self.headers.update(user_agent)

        # configure proxy
        if proxies and proxies_enabled:
            self.proxies = proxies

        # configure verify
        self.verify = verify

        # Add Retry
        self.retry()

    def _log_curl(self, response: Response):
        """Log the curl equivalent command."""

        # don't show curl message for logging commands
        # APP-79 - adding logging of request as curl commands
        # if response.request.url is not None and '/v2/logs/app' not in response.request.url:
        #     if not response.ok or self.log_curl:
        if (response.request.url is not None and '/v2/logs/app' not in response.request.url) and (
            not response.ok or self.log_curl
        ):
            with contextlib.suppress(Exception):
                self.log.debug(
                    self.requests_to_curl.convert(
                        response.request, proxies=self.proxies, verify=self.verify
                    )
                )

    def request(self, method, url, **kwargs):  # type: ignore
        """Override request method disabling verify on token renewal if disabled on session."""
        response = super().request(method, self.url(url), **kwargs)
        bad_request_code = 401

        # retry request in case we encountered a race condition with token renewal monitor
        if response.status_code == bad_request_code:
            self.log.debug(
                f'Unexpected response received while attempting to send a request using internal '
                f'session object. Retrying request. feature=tc-session, '
                f'request-url={response.request.url}, status-code={response.status_code}'
            )
            response = super().request(method, self.url(url), **kwargs)

        # optionally log the curl command
        self._log_curl(response)

        # log request and response data
        self.log.debug(
            f'feature=tc-session, method={method}, request-url={response.request.url}, '
            f'status-code={response.status_code}, elapsed={response.elapsed}'
        )

        return response

    def retry(self, retries=3, backoff_factor=0.3, status_forcelist=(500, 502, 504)):
        """Add retry to Requests Session

        https://urllib3.readthedocs.io/en/latest/reference/urllib3.util.html#urllib3.util.retry.Retry
        """
        retries = Retry(
            total=retries,
            read=retries,
            connect=retries,
            backoff_factor=backoff_factor,  # type: ignore
            status_forcelist=status_forcelist,
        )
        # mount all https requests
        self.mount('https://', adapters.HTTPAdapter(max_retries=retries))

    def url(self, url: str) -> str:
        """Return appropriate URL string.

        The method allows the session to accept the URL Path or the full URL.
        """
        if not url.startswith('https'):
            return f'{self.base_url}{url}'
        return url



"""TcEx Framework Module"""

# standard library
from datetime import datetime

# third-party
from pydantic import BaseModel, Extra, Field, PrivateAttr, validator

# first-party
from tcex.api.tc.v3.v3_model_abc import V3ModelABC
from tcex.util import Util


class GroupModel(
    V3ModelABC,
    alias_generator=Util().snake_to_camel,
    extra=Extra.allow,
    title='Group Model',
    validate_assignment=True,
):
    """Group Model"""

    _associated_type = PrivateAttr(default=True)
    _cm_type = PrivateAttr(default=False)
    _shared_type = PrivateAttr(default=False)
    _staged = PrivateAttr(default=False)

    assignments: 'TaskAssigneesModel' = Field(
        None,
        description=(
            'A list of assignees and escalatees associated with this group (Task specific).'
        ),
        methods=['POST', 'PUT'],
        read_only=False,
        title='assignments',
    )
    associated_artifacts: 'ArtifactsModel' = Field(
        None,
        description='A list of Artifacts associated with this Group.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='associatedArtifacts',
    )
    associated_cases: 'CasesModel' = Field(
        None,
        description='A list of Cases associated with this Group.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='associatedCases',
    )
    associated_groups: 'GroupsModel' = Field(
        None,
        description='A list of groups associated with this group.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='associatedGroups',
    )
    associated_indicators: 'IndicatorsModel' = Field(
        None,
        description='A list of indicators associated with this group.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='associatedIndicators',
    )
    associated_victim_assets: 'VictimAssetsModel' = Field(
        None,
        description='A list of victim assets associated with this group.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='associatedVictimAssets',
    )
    attributes: 'GroupAttributesModel' = Field(
        None,
        description='A list of Attributes corresponding to the Group.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='attributes',
    )
    body: str | None = Field(
        None,
        applies_to=['Email'],
        description='The email Body.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='body',
    )
    created_by: 'UserModel' = Field(
        None,
        allow_mutation=False,
        description='The **created by** for the Group.',
        read_only=True,
        title='createdBy',
    )
    date_added: datetime | None = Field(
        None,
        allow_mutation=False,
        description='The date and time that the item was first created.',
        read_only=True,
        title='dateAdded',
    )
    document_date_added: datetime | None = Field(
        None,
        allow_mutation=False,
        applies_to=['Document', 'Report'],
        description='The date and time that the document was first created.',
        read_only=True,
        title='documentDateAdded',
    )
    document_type: str | None = Field(
        None,
        allow_mutation=False,
        applies_to=['Document', 'Report'],
        description='The document type.',
        read_only=True,
        title='documentType',
    )
    down_vote_count: int | None = Field(
        None,
        allow_mutation=False,
        description='The total number of users who find the intel not helpful.',
        read_only=True,
        title='downVoteCount',
    )
    due_date: datetime | None = Field(
        None,
        applies_to=['Task'],
        description='The date and time that the Task is due.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='dueDate',
    )
    email_date: datetime | None = Field(
        None,
        allow_mutation=False,
        applies_to=['Email'],
        description='The date and time that the email was first created.',
        read_only=True,
        title='emailDate',
    )
    escalated: bool = Field(
        None,
        allow_mutation=False,
        applies_to=['Task'],
        description='Flag indicating whether or not the task has been escalated.',
        read_only=True,
        title='escalated',
    )
    escalation_date: datetime | None = Field(
        None,
        applies_to=['Task'],
        description='The escalation date and time.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='escalationDate',
    )
    event_date: datetime | None = Field(
        None,
        applies_to=['Incident', 'Event'],
        description='The date and time that the incident or event was first created.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='eventDate',
    )
    event_type: str | None = Field(
        None,
        applies_to=['Event'],
        description='The identification of an event type.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='eventType',
    )
    external_date_added: datetime | None = Field(
        None,
        description='The date and time that the item was first created externally.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='externalDateAdded',
    )
    external_date_expires: datetime | None = Field(
        None,
        description='The date and time the item expires externally.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='externalDateExpires',
    )
    external_last_modified: datetime | None = Field(
        None,
        description='The date and time the item was modified externally.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='externalLastModified',
    )
    file_name: str | None = Field(
        None,
        applies_to=['Document', 'Report', 'Signature'],
        conditional_required=['Document', 'Report', 'Signature'],
        description='The document or signature file name.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='fileName',
    )
    file_size: int | None = Field(
        None,
        allow_mutation=False,
        applies_to=['Document', 'Report'],
        description='The document file size.',
        read_only=True,
        title='fileSize',
    )
    file_text: str | None = Field(
        None,
        applies_to=['Signature'],
        description='The signature file text.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='fileText',
    )
    file_type: str | None = Field(
        None,
        applies_to=['Signature'],
        description='The signature file type.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='fileType',
    )
    first_seen: datetime | None = Field(
        None,
        description='The date and time that the item was first seen.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='firstSeen',
    )
    from_: str | None = Field(
        None,
        alias='from',
        applies_to=['Email'],
        description='The email From field.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='from',
    )
    generated_report: bool = Field(
        None,
        allow_mutation=False,
        description='Is the report auto-generated?',
        read_only=True,
        title='generatedReport',
    )
    header: str | None = Field(
        None,
        applies_to=['Email'],
        description='The email Header field.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='header',
    )
    id: int | None = Field(  # type: ignore
        None,
        description='The ID of the item.',
        read_only=True,
        title='id',
    )
    insights: dict | None = Field(
        None,
        allow_mutation=False,
        applies_to=['Document', 'Report'],
        description='An AI generated synopsis of the document.',
        read_only=True,
        title='insights',
    )
    last_modified: datetime | None = Field(
        None,
        allow_mutation=False,
        description='The date and time that the Entity was last modified.',
        read_only=True,
        title='lastModified',
    )
    last_seen: datetime | None = Field(
        None,
        description='The date and time that the item was last seen.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='lastSeen',
    )
    legacy_link: str | None = Field(
        None,
        allow_mutation=False,
        description='A link to the legacy ThreatConnect details page for this entity.',
        read_only=True,
        title='legacyLink',
    )
    malware: bool = Field(
        None,
        applies_to=['Document'],
        description='Is the document malware?',
        methods=['POST', 'PUT'],
        read_only=False,
        title='malware',
    )
    name: str | None = Field(
        None,
        description='The name of the group.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='name',
    )
    overdue: bool = Field(
        None,
        allow_mutation=False,
        applies_to=['Task'],
        description='Flag indicating whether or not the task is overdue.',
        read_only=True,
        title='overdue',
    )
    owner_id: int | None = Field(
        None,
        description='The id of the Organization, Community, or Source that the item belongs to.',
        methods=['POST'],
        read_only=False,
        title='ownerId',
    )
    owner_name: str | None = Field(
        None,
        conditional_read_only=['Victim'],
        description='The name of the Organization, Community, or Source that the item belongs to.',
        methods=['POST'],
        read_only=False,
        title='ownerName',
    )
    password: str | None = Field(
        None,
        applies_to=['Document'],
        description='The password associated with the document (Required if Malware is true).',
        methods=['POST', 'PUT'],
        read_only=False,
        title='password',
    )
    publish_date: datetime | None = Field(
        None,
        applies_to=['Report'],
        description='The date and time that the report was first created.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='publishDate',
    )
    reminded: bool = Field(
        None,
        allow_mutation=False,
        applies_to=['Task'],
        description='Flag indicating whether or not the task reminders have been sent.',
        read_only=True,
        title='reminded',
    )
    reminder_date: datetime | None = Field(
        None,
        applies_to=['Task'],
        description='The reminder date and time.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='reminderDate',
    )
    reviews: dict | None = Field(
        None,
        description='A list of reviews corresponding to the Group.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='reviews',
    )
    score: int | None = Field(
        None,
        allow_mutation=False,
        applies_to=['Email'],
        description='The score value for this email.',
        read_only=True,
        title='score',
    )
    score_breakdown: str | None = Field(
        None,
        allow_mutation=False,
        applies_to=['Email'],
        description='The email score breakdown.',
        read_only=True,
        title='scoreBreakdown',
    )
    score_includes_body: bool = Field(
        None,
        allow_mutation=False,
        applies_to=['Email'],
        description='Is the Body included in the email score?',
        read_only=True,
        title='scoreIncludesBody',
    )
    security_labels: 'SecurityLabelsModel' = Field(
        None,
        description=(
            'A list of Security Labels corresponding to the Intel item (NOTE: Setting this '
            'parameter will replace any existing tag(s) with the one(s) specified).'
        ),
        methods=['POST', 'PUT'],
        read_only=False,
        title='securityLabels',
    )
    signature_date_added: datetime | None = Field(
        None,
        allow_mutation=False,
        applies_to=['Signature'],
        description='The date and time that the signature was first created.',
        read_only=True,
        title='signatureDateAdded',
    )
    status: str | None = Field(
        None,
        applies_to=['Document', 'Report', 'Event', 'Task', 'Incident'],
        conditional_read_only=['Document', 'Report', 'Task'],
        description=(
            'The status associated with this document, event, task, or incident (read only for '
            'task, document, and report).'
        ),
        methods=['POST', 'PUT'],
        read_only=False,
        title='status',
    )
    subject: str | None = Field(
        None,
        applies_to=['Email'],
        description='The email Subject section.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='subject',
    )
    tags: 'TagsModel' = Field(
        None,
        description=(
            'A list of Tags corresponding to the item (NOTE: Setting this parameter will replace '
            'any existing tag(s) with the one(s) specified).'
        ),
        methods=['POST', 'PUT'],
        read_only=False,
        title='tags',
    )
    to: str | None = Field(
        None,
        allow_mutation=False,
        applies_to=['Email'],
        description='The email To field .',
        read_only=True,
        title='to',
    )
    type: str | None = Field(
        None,
        description='The **type** for the Group.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='type',
    )
    up_vote: bool = Field(
        None,
        allow_mutation=False,
        description=(
            'Is the intelligence valid and useful? (0 means downvote, 1 means upvote, and NULL '
            'means no vote).'
        ),
        read_only=True,
        title='upVote',
    )
    up_vote_count: int | None = Field(
        None,
        allow_mutation=False,
        description='The total number of users who find the intel useful.',
        read_only=True,
        title='upVoteCount',
    )
    web_link: str | None = Field(
        None,
        allow_mutation=False,
        description='A link to the ThreatConnect details page for this entity.',
        read_only=True,
        title='webLink',
    )
    xid: str | None = Field(
        None,
        description='The xid of the item.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='xid',
    )

    @validator('associated_artifacts', always=True, pre=True)
    @classmethod
    def _validate_artifacts(cls, v):
        if not v:
            return ArtifactsModel()  # type: ignore
        return v

    @validator('associated_cases', always=True, pre=True)
    @classmethod
    def _validate_cases(cls, v):
        if not v:
            return CasesModel()  # type: ignore
        return v

    @validator('attributes', always=True, pre=True)
    @classmethod
    def _validate_group_attributes(cls, v):
        if not v:
            return GroupAttributesModel()  # type: ignore
        return v

    @validator('associated_groups', always=True, pre=True)
    @classmethod
    def _validate_groups(cls, v):
        if not v:
            return GroupsModel()  # type: ignore
        return v

    @validator('associated_indicators', always=True, pre=True)
    @classmethod
    def _validate_indicators(cls, v):
        if not v:
            return IndicatorsModel()  # type: ignore
        return v

    @validator('security_labels', always=True, pre=True)
    @classmethod
    def _validate_security_labels(cls, v):
        if not v:
            return SecurityLabelsModel()  # type: ignore
        return v

    @validator('tags', always=True, pre=True)
    @classmethod
    def _validate_tags(cls, v):
        if not v:
            return TagsModel()  # type: ignore
        return v

    @validator('assignments', always=True, pre=True)
    @classmethod
    def _validate_task_assignees(cls, v):
        if not v:
            return TaskAssigneesModel()  # type: ignore
        return v

    @validator('created_by', always=True, pre=True)
    @classmethod
    def _validate_user(cls, v):
        if not v:
            return UserModel()  # type: ignore
        return v

    @validator('associated_victim_assets', always=True, pre=True)
    @classmethod
    def _validate_victim_assets(cls, v):
        if not v:
            return VictimAssetsModel()  # type: ignore
        return v


class GroupDataModel(
    BaseModel,
    title='Group Data Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Groups Data Model"""

    data: list[GroupModel] | None = Field(
        [],
        description='The data for the Groups.',
        methods=['POST', 'PUT'],
        title='data',
    )


class GroupsModel(
    BaseModel,
    title='Groups Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Groups Model"""

    _mode_support = PrivateAttr(default=True)

    data: list[GroupModel] | None = Field(
        [],
        description='The data for the Groups.',
        methods=['POST', 'PUT'],
        title='data',
    )
    mode: str = Field(
        'append',
        description='The PUT mode for nested objects (append, delete, replace). Default: append',
        methods=['POST', 'PUT'],
        title='append',
    )


# first-party
from tcex.api.tc.v3.artifacts.artifact_model import ArtifactsModel
from tcex.api.tc.v3.cases.case_model import CasesModel
from tcex.api.tc.v3.group_attributes.group_attribute_model import GroupAttributesModel
from tcex.api.tc.v3.indicators.indicator_model import IndicatorsModel
from tcex.api.tc.v3.security.task_assignee_model import TaskAssigneesModel
from tcex.api.tc.v3.security.users.user_model import UserModel
from tcex.api.tc.v3.security_labels.security_label_model import SecurityLabelsModel
from tcex.api.tc.v3.tags.tag_model import TagsModel
from tcex.api.tc.v3.victim_assets.victim_asset_model import VictimAssetsModel

# add forward references
GroupDataModel.update_forward_refs()
GroupModel.update_forward_refs()
GroupsModel.update_forward_refs()



"""TcEx Framework Module"""

# third-party
from pydantic import BaseModel, Extra, Field, PrivateAttr

# first-party
from tcex.api.tc.v3.v3_model_abc import V3ModelABC
from tcex.util import Util


class UserModel(
    V3ModelABC,
    alias_generator=Util().snake_to_camel,
    extra=Extra.allow,
    title='User Model',
    validate_assignment=True,
):
    """User Model"""

    _associated_type = PrivateAttr(default=False)
    _cm_type = PrivateAttr(default=False)
    _shared_type = PrivateAttr(default=False)
    _staged = PrivateAttr(default=False)

    id: int | None = Field(  # type: ignore
        None,
        description='The ID of the item.',
        read_only=True,
        title='id',
    )
    user_name: str | None = Field(
        None,
        allow_mutation=False,
        description='The **user name** for the User.',
        read_only=True,
        title='userName',
    )


class UserDataModel(
    BaseModel,
    title='User Data Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Users Data Model"""

    data: list[UserModel] | None = Field(
        [],
        description='The data for the Users.',
        methods=['POST', 'PUT'],
        title='data',
    )


class UsersModel(
    BaseModel,
    title='Users Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Users Model"""

    _mode_support = PrivateAttr(default=False)

    data: list[UserModel] | None = Field(
        [],
        description='The data for the Users.',
        methods=['POST', 'PUT'],
        title='data',
    )
    mode: str = Field(
        'append',
        description='The PUT mode for nested objects (append, delete, replace). Default: append',
        methods=['POST', 'PUT'],
        title='append',
    )


# add forward references
UserDataModel.update_forward_refs()
UserModel.update_forward_refs()
UsersModel.update_forward_refs()



"""TcEx Framework Module"""

# standard library
from typing import TYPE_CHECKING, Any

# third-party
import wrapt

if TYPE_CHECKING:
    # standard library
    from collections.abc import Callable


class Output:
    """Store the method return value in self.<attribute>.

    This decorator will write, append, or extend the methods return value to the App attribute
    provided in the ``attribute`` input. The ``attribute`` must first be defined in the
    ``__init__()`` method of the App before the decorator is used.

    .. code-block:: python
        :linenos:
        :lineno-start: 1

        def __init__(self, _tcex):
            super(App, self).__init__(_tcex)
            self.output_strings = []  # Output decorator writes here.


        @Output(attribute='output_strings')
        def my_method(data):
            return data.lowercase()

    Args:
        attribute: The name of the App attribute to write data.
        overwrite: When True and the method is called more than once the previous value
            will be overwritten.
    """

    def __init__(self, attribute: str, overwrite: bool = False):
        """Initialize instance properties"""
        self.attribute = attribute
        self.overwrite = overwrite

    @wrapt.decorator
    def __call__(self, *wrapped_args) -> Any:  # noqa: D417
        """Implement __call__ function for decorator.

        Args:
            wrapped: The wrapped function which in turns
                needs to be called by your wrapper function.
            instance: The object to which the wrapped
                function was bound when it was called.
            args: The list of positional arguments supplied
                when the decorated function was called.
            kwargs: The dictionary of keyword arguments
                supplied when the decorated function was called.
        """
        # using wrapped args to support typing hints in PyRight
        wrapped: Callable = wrapped_args[0]
        app: Any = wrapped_args[1]
        args: list = wrapped_args[2] if len(wrapped_args) > 1 else []
        kwargs: dict = wrapped_args[3] if len(wrapped_args) > 2 else {}  # noqa: PLR2004

        def output() -> Any:
            """Call the function and store or append return value.

            Args:
                app (class): The instance of the App class "self".
                *args: Additional positional arguments.
                **kwargs: Additional keyword arguments.
            """
            data = wrapped(*args, **kwargs)
            attr = getattr(app, self.attribute)

            # tracker to indicate see if attribute has already been updated
            attr_tracker_name = f'__{self.attribute}_tracker__'
            attr_tracker = False
            try:
                attr_tracker = getattr(app, attr_tracker_name)
            except AttributeError:
                setattr(app, attr_tracker_name, False)

            # update the attribute
            if isinstance(attr, list) and isinstance(data, list) and not self.overwrite:
                attr.extend(data)
            elif isinstance(attr, list) and not self.overwrite:
                attr.append(data)
            elif attr_tracker and isinstance(data, list) and not self.overwrite:
                # convert string to list and extend with data
                d = [attr]
                d.extend(data)
                setattr(app, self.attribute, d)
            elif attr_tracker and not self.overwrite:
                # convert string to list and append with data
                d = [attr]
                d.append(data)
                setattr(app, self.attribute, d)
            else:
                setattr(app, self.attribute, data)

            # update tracker to indicate the attribute has already been set at least once
            setattr(app, attr_tracker_name, True)

            return data

        return output()



"""TcEx Framework Module"""

# standard library
from abc import ABC
from collections.abc import Generator

# first-party
from tcex.api.tc.v3._gen._gen_abc import GenerateABC
from tcex.api.tc.v3._gen._gen_args_abc import GenerateArgsABC
from tcex.api.tc.v3._gen._options_data_object import OptionsDataObject
from tcex.api.tc.v3._gen.model import PropertyModel
from tcex.util.string_operation import SnakeString


class GenerateArgs(GenerateArgsABC):
    """Generate Models for TC API Types"""


class GenerateObjectABC(GenerateABC, ABC):
    """Generate Models for Case Management Types"""

    def __init__(self, type_: SnakeString):
        """Initialize instance properties."""
        super().__init__(type_)

        # properties
        self.requirements = {
            'standard library': [],
            'third-party': [],
            'first-party': [
                'from tcex.api.tc.v3.api_endpoints import ApiEndpoints',
                'from tcex.api.tc.v3.object_abc import ObjectABC',
                'from tcex.api.tc.v3.object_collection_abc import ObjectCollectionABC',
            ],
            'first-party-forward-reference': [],
            'type-checking': [],
        }

    @property
    def content_models(self) -> Generator[PropertyModel, None, None]:
        """Return the options for the current type."""
        yield from self.options_data.content_models
        self.messages.extend(self.options_data.messages)

    @property
    def options_data(self) -> OptionsDataObject:
        """Return the download options for the current type."""
        return OptionsDataObject(api_url=self.api_url, object_type=self.type_)

    def _gen_code_api_endpoint_property(self) -> str:
        """Return the method code.

        @property
        def _api_endpoint(self) -> str:
            '''Return the type specific API endpoint.'''
            return ApiEndpoints.ARTIFACTS.value
        """
        return '\n'.join(
            [
                f'{self.i1}@property',
                f'{self.i1}def _api_endpoint(self) -> str:',
                f'{self.i2}"""Return the type specific API endpoint."""',
                f'{self.i2}return ApiEndpoints.{self.type_.upper()}.value',
                '',
                '',
            ]
        )

    def _gen_code_container_init_method(self) -> str:
        """Return the method code.

        def __init__(self, **kwargs):
            '''Initialize instance properties.'''
            super().__init__(
                kwargs.pop('session', None),
                kwargs.pop('tql_filter', None),
                kwargs.pop('params', None)
            )
            self._model = ArtifactsModel(**kwargs)
        """
        # add method import requirement
        classes = [
            f'{self.type_.singular().pascal_case()}Model',
            f'{self.type_.plural().pascal_case()}Model',
        ]
        self.update_requirements(self.type_, f'{self.type_.singular()}_model', classes)

        model_name = f'{self.type_.plural().pascal_case()}Model'
        return '\n'.join(
            [
                f'{self.i1}def __init__(self, **kwargs):',
                f'{self.i2}"""Initialize instance properties."""',
                f'{self.i2}super().__init__(',
                (
                    f'{self.i3}kwargs.pop("session", None), '
                    f'kwargs.pop("tql_filter", None), '
                    f'kwargs.pop("params", None)'
                ),
                f'{self.i2})',
                f'{self.i2}self._model = {model_name}(**kwargs)',
                f'{self.i2}self.type_ = "{self.type_.plural()}"',
                '',
                '',
            ]
        )

    def _gen_code_container_iter_method(self) -> str:
        """Return the method code.

        def __iter__(self) -> Iterator[Artifact]:
            '''Return CM objects.'''
            return self.iterate(base_class=Artifact)
        """
        self.requirements['standard library'].append('from collections.abc import Iterator')
        return '\n'.join(
            [
                (
                    f'{self.i1}def __iter__(self) -> '
                    f'Iterator[{self.type_.singular().pascal_case()}]:'
                ),
                f'{self.i2}"""Return CM objects."""',
                (
                    f'{self.i2}return self.iterate(base_class='
                    f'{self.type_.singular().pascal_case()})'
                    '  # type: ignore'
                ),
                '',
                '',
            ]
        )

    # def _gen_code_container_iter_method(self) -> str:
    #     """Return the method code.

    #     def __iter__(self) -> Self:
    #         '''Return CM objects.'''
    #         return self
    #     """
    #     self.requirements['standard library'].append({'module': 'typing', 'imports': ['Self']})
    #     return '\n'.join(
    #         [
    #             f'''{self.i1}def __iter__(self) -> Self:''',
    #             f'''{self.i2}"""Return CM objects."""''',
    #             f'''{self.i2}return self''',
    #             '',
    #             '',
    #         ]
    #     )

    # def _gen_code_container_next_method(self) -> str:
    #     """Return the method code.

    #     def __next__(self) -> Artifact:
    #         '''Iterate over CM objects.'''
    #         return self.iterate(base_class=Artifact)
    #     """
    #     return '\n'.join(
    #         [
    #             f'''{self.i1}def __next__(self) -> {self.type_.singular().pascal_case()}:''',
    #             f'''{self.i2}"""Return next CM objects."""''',
    #             (
    #                 f'''{self.i2}for i in self.iterate(base_class='''
    #                 f'''{self.type_.singular().pascal_case()}):'''
    #             ),
    #             f'''{self.i3}return i''',
    #             '',
    #             f'''{self.i2}raise StopIteration''',
    #             '',
    #             '',
    #         ]
    #     )

    def _gen_code_container_filter_property(self) -> str:
        """Return the method code.

        @property
        def filter(self) -> ArtifactFilter:
            '''Return the type specific filter object.'''
            return ArtifactFilter(self._session, self.tql)
        """
        filter_class = f'{self.type_.singular().pascal_case()}Filter'
        self.update_requirements(self.type_, f'{self.type_.singular()}_filter', [filter_class])
        return '\n'.join(
            [
                f'{self.i1}@property',
                f'{self.i1}def filter(self) -> {filter_class}:',
                f'{self.i2}"""Return the type specific filter object."""',
                f'{self.i2}return {filter_class}(self.tql)',
                '',
                '',
            ]
        )

    def _gen_code_deleted_method(self) -> str:
        """Return the method code.

        @property
        def filter(self) -> ArtifactFilter:
            '''Return the type specific filter object.'''
            return ArtifactFilter(self._session, self.tql)
        """
        self.requirements['first-party'].append({'module': 'datetime', 'imports': ['datetime']})
        return '\n'.join(
            [
                f'{self.i1}def deleted(',
                f'{self.i2}self,',
                f'{self.i2}deleted_since: datetime | str | None,',
                f'{self.i2}type_: str | None = None,',
                f'{self.i2}owner: str | None = None',
                f'{self.i1}):',
                f'{self.i2}"""Return deleted indicators.',
                '',
                f'{self.i2}This will not use the default params set on the "Indicators" ',
                f'{self.i2}object and instead used the params that are passed in.',
                f'{self.i2}"""',
                '',
                f'{self.i2}if deleted_since is not None:',
                f'{self.i3}deleted_since = str(',
                (
                    f'{self.i4}self.util.any_to_datetime(deleted_since)'
                    '.strftime("%Y-%m-%dT%H:%M:%SZ")'
                ),
                f'{self.i3})',
                '',
                f'{self.i2}yield from self.iterate(',
                f'{self.i3}base_class=Indicator,',
                f'{self.i3}api_endpoint=f"{{self._api_endpoint}}/deleted",',
                (
                    f'{self.i3}params={{"deletedSince": deleted_since, '
                    '"owner": owner, "type": type_}'
                ),
                f'{self.i2})',
                '',
            ]
        )

    def _gen_code_group_methods(self) -> str:
        """Return the method code.

        @property
        def filter(self) -> ArtifactFilter:
            '''Return the type specific filter object.'''
            return ArtifactFilter(self._session, self.tql)
        """
        # self.requirements['type-checking'].append('''from requests import Response''')
        self.requirements['first-party'].append("""from requests import Response""")
        return '\n'.join(
            [
                f'{self.i1}def download(self, params: dict | None = None) -> bytes:',
                f'{self.i2}"""Return the document attachment for Document/Report Types."""',
                f'{self.i2}self._request(',
                f'{self.i3}method="GET",',
                f'{self.i3}url=f\'{{self.url("GET")}}/download\',',
                f'{self.i3}headers={{"Accept": "application/octet-stream"}},',
                f'{self.i3}params=params,',
                f'{self.i2})',
                f'{self.i2}return self.request.content',
                '',
                f'{self.i1}def pdf(self, params: dict | None = None) -> bytes:',
                f'{self.i2}"""Return the document attachment for Document/Report Types."""',
                f'{self.i2}self._request(',
                f'{self.i3}method="GET",',
                f'{self.i3}body=None,',
                f'{self.i3}url=f\'{{self.url("GET")}}/pdf\',',
                f'{self.i3}headers={{"Accept": "application/octet-stream"}},',
                f'{self.i3}params=params,',
                f'{self.i2})',
                '',
                f'{self.i2}return self.request.content',
                '',
                (
                    f'{self.i1}def upload(self, content: bytes | str, '
                    'params: dict | None = None) -> Response:'
                ),
                f'{self.i2}"""Return the document attachment for Document/Report Types."""',
                f'{self.i2}self._request(',
                f'{self.i3}method="POST",',
                f'{self.i3}url=f\'{{self.url("GET")}}/upload\',',
                f'{self.i3}body=content,',
                f'{self.i3}headers={{"content-type": "application/octet-stream"}},',
                f'{self.i3}params=params,',
                f'{self.i2})',
                f'{self.i2}return self.request',
                '',
                '',
            ]
        )

    def _gen_code_object_init_method(self) -> str:
        """Return the method code.

        def __init__(self, **kwargs):
            '''Initialize instance properties'''
            super().__init__(kwargs.pop('session', None))
            self._model = ArtifactModel(**kwargs)
        """
        # set nested type
        nested_field_name = self.util.snake_string(self.type_).camel_case().plural()
        if self.type_ in ['indicators', 'groups']:
            nested_field_name = (
                self.util.snake_string(f'associated_{self.type_}').camel_case().plural()
            )
        elif self.type_ in ['case_attributes', 'group_attributes', 'indicator_attributes']:
            nested_field_name = 'attributes'

        model_name = f'{self.type_.singular().pascal_case()}Model'
        return '\n'.join(
            [
                f'{self.i1}def __init__(self, **kwargs):',
                f'{self.i2}"""Initialize instance properties."""',
                f'{self.i2}super().__init__(kwargs.pop("session", None))',
                '',
                f'{self.i2}# properties',
                f'{self.i2}self._model: {model_name} = {model_name}(**kwargs)',
                f'{self.i2}self._nested_field_name = "{nested_field_name}" ',
                f'{self.i2}self._nested_filter = "has_{self.type_.singular()}" ',
                f'{self.i2}self.type_ = "{self.type_.singular().space_case()}"',
                '',
                '',
            ]
        )

    def _gen_code_object_model_property(self) -> str:
        """Return the method code.

        Override the object ABC method so that the correct typing hint return
        type can be used and the IDE will provide full model hinting.

        @property
        def model(self) -> CaseModel:
            '''Return the model data.'''
            return self._model

        @model.setter
        def model(self, data: dict | IndicatorModel):
            '''Create model using the provided data.'''
            if isinstance(data, type(self.model)):
                # provided data is already a model, nothing required to change
                self._model = data
            elif isinstance(data, dict):
                # provided data is raw response, load the model
                self._model = type(self.model)(**data)
            else:
                ex_msg = f'Invalid data type: {type(data)} provided.'
                raise RuntimeError(ex_msg)
        """
        return '\n'.join(
            [
                f'{self.i1}@property',
                f'{self.i1}def model(self) -> {self.type_.singular().pascal_case()}Model:',
                f'{self.i2}"""Return the model data."""',
                f'{self.i2}return self._model',
                '',
                f'{self.i1}@model.setter',
                (
                    f'{self.i1}def model(self, data: '
                    f'dict | {self.type_.singular().pascal_case()}Model):'
                ),
                f'{self.i2}"""Create model using the provided data."""',
                f'{self.i2}if isinstance(data, type(self.model)):',
                f'{self.i3}# provided data is already a model, nothing required to change',
                f'{self.i3}self._model = data',
                f'{self.i2}elif isinstance(data, dict):',
                f'{self.i3}# provided data is raw response, load the model',
                f'{self.i3}self._model = type(self.model)(**data)',
                f'{self.i2}else:',
                f'{self.i3}ex_msg = f"Invalid data type: {{type(data)}} provided."',
                f'{self.i3}raise RuntimeError(ex_msg)  # noqa: TRY004',
                '',
                '',
            ]
        )

    def _gen_code_object_as_entity_property_method(self) -> str:
        """Return the method code.

        @property
        def as_entity(self) -> dict:
            '''Return the entity representation of the object.'''
            case - name
            notes - summary
            tag - name
            task - name
            workflow_event - summary
            workflow_template_model - name

            return {'type': 'Artifact', 'id': self.model.id, 'value': self.model.summary}
        """
        name_entities = [
            'artifact_types',
            'cases',
            'categories',
            'results',
            'subtypes',
            'tags',
            'tasks',
            'workflow_templates',
            'groups',
            'victims',
        ]
        # check if groups or indicators and if so use the self.model.type else use self.type_
        value = 'self.model.summary'
        if self.type_.lower() in name_entities:
            value = 'self.model.name'
        elif self.type_.lower() == 'intel_requirements':
            value = 'self.model.requirement_text'

        as_entity_property_method = [
            f'{self.i1}@property',
            f'{self.i1}def as_entity(self) -> dict:',
            f'{self.i2}"""Return the entity representation of the object."""',
        ]
        if self.type_.lower() in ['groups', 'indicators']:
            as_entity_property_method.append(f'{self.i2}type_ = self.model.type')
        elif self.type_.lower() in ['victim_assets']:
            value = 'value'
            as_entity_property_method.extend(
                [
                    f'{self.i2}value = []',
                    '',
                    f'{self.i2}if self.model.type is not None:',
                    f'{self.i3}if self.model.type.lower() == "phone":',
                    f'{self.i4}if self.model.phone:',
                    f'{self.i5}value.append(self.model.phone)',
                    f'{self.i3}elif self.model.type.lower() == "socialnetwork":',
                    f'{self.i4}if self.model.social_network:',
                    f'{self.i5}value.append(self.model.social_network)',
                    f'{self.i4}if self.model.account_name:',
                    f'{self.i5}value.append(self.model.account_name)',
                    f'{self.i3}elif self.model.type.lower() == "networkaccount":',
                    f'{self.i4}if self.model.network_type:',
                    f'{self.i5}value.append(self.model.network_type)',
                    f'{self.i4}if self.model.account_name:',
                    f'{self.i5}value.append(self.model.account_name)',
                    f'{self.i3}elif self.model.type.lower() == "emailaddress":',
                    f'{self.i4}if self.model.address_type:',
                    f'{self.i5}value.append(self.model.address_type)',
                    f'{self.i4}if self.model.address:',
                    f'{self.i5}value.append(self.model.address)',
                    f'{self.i3}elif self.model.type.lower() == "website" and self.model.website:',
                    f'{self.i4}value.append(self.model.website)',
                    '',
                    '',
                    f'{self.i2}value = " : ".join(value) if value else ""',
                    f'{self.i2}type_ = f"Victim Asset : {{self.model.type}}"',
                ]
            )
        else:
            as_entity_property_method.append(f'{self.i2}type_ = self.type_')
        as_entity_property_method.extend(
            [
                '',
                f'{self.i2}return {{"type": type_, "id": self.model.id, "value": {value}}}',
                '',
                '',
            ]
        )
        return '\n'.join(as_entity_property_method)

    def _gen_code_object_replace_type_method(
        self, type_: str, model_type: str | None = None
    ) -> str:
        """Return the method code.

        def replace_artifact(self, **kwargs):
            '''Replace an Artifact on the object. (mark as staged)
        """
        type_ = self.util.snake_string(type_)
        model_type = self.util.snake_string(model_type or type_)
        model_reference = model_type

        # get model from map and update requirements
        model_import_data = self._module_import_data(type_)
        model_class = model_import_data.get('model_class')
        self.requirements['first-party'].append(
            f'from {model_import_data.get("model_module")} '
            f'import {model_import_data.get("model_class")}'
        )
        stage_method = [
            (
                f'{self.i1}def replace_{model_type.singular()}(self, '
                f'data: dict | list | ObjectABC | {model_class}'
                f'):'
            ),
            f'{self.i2}"""Replace {type_.singular()} on the object."""',
            f'{self.i2}if not isinstance(data, list):',
            f'{self.i3}data = [data]',
            '',
            f'{self.i2}if all(isinstance(item, ({model_class} | ObjectABC)) for item in data):'
            f'{self.i3}transformed_data = data',
            f'{self.i2}elif all(isinstance(item, dict) for item in data):'
            f'{self.i3}transformed_data = [{model_class}(**d) for d in data]',
            f'{self.i2}else:',
            f'{self.i3}ex_msg = "Invalid data to replace_{model_type.singular()}"',
            f'{self.i3}raise ValueError(ex_msg)',
            '',
            '',
            f'{self.i2}for item in transformed_data:',
            f'{self.i3}item._staged = True  # noqa: SLF001',
            '',
            f'{self.i2}self.model.{model_reference} = transformed_data  # type: ignore',
            '',
            '',
        ]

        return '\n'.join(stage_method)

    def _gen_code_object_stage_type_method(self, type_: str, model_type: str | None = None) -> str:
        """Return the method code.

        def stage_artifact(self, **kwargs):
            '''Stage an Artifact on the object.

            ...
            '''
            self.model.artifacts.data.append(ArtifactModel(**kwargs))

            _code += self._gen_code_object_add_type_method('users', 'user_access')
        """
        type_ = self.util.snake_string(type_)
        model_type = self.util.snake_string(model_type or type_)
        model_reference = model_type

        # Unlike all of the other objects, on the victims model, it references 'assets' not the
        # model name 'VictimAssets'
        if type_.lower() == 'victim_assets' and self.type_.lower() == 'victims':
            model_reference = self.util.camel_string('assets')
        elif type_.lower() == 'users':
            model_type = self.util.camel_string('user_accesses')
            model_reference = 'user_access'

        # get model from map and update requirements
        model_import_data = self._module_import_data(type_)
        self.requirements['first-party'].append(
            f'from {model_import_data.get("model_module")} '
            f'import {model_import_data.get("model_class")}'
        )
        stage_method = [
            (
                f'{self.i1}def stage_{model_type.singular()}(self, '
                f'data: dict | ObjectABC | {model_import_data.get("model_class")}'
                f'):'
            ),
            f'{self.i2}"""Stage {type_.singular()} on the object."""',
            f'{self.i2}if isinstance(data, ObjectABC):',
            f'{self.i3}data = data.model  # type: ignore',
            f'{self.i2}elif isinstance(data, dict):',
            f'{self.i3}data = {model_import_data.get("model_class")}(**data)',
            '',
            f'{self.i2}if not isinstance(data, {model_import_data.get("model_class")}):',
            f'{self.i3}ex_msg = "Invalid type passed in to stage_{model_type.singular()}"',
            f'{self.i3}raise RuntimeError(ex_msg)  # noqa: TRY004',
            f'{self.i2}data._staged = True  # noqa: SLF001',
        ]
        if type_.lower() == 'file_actions' and self.type_.lower() == 'indicators':
            # The `indicator` field in the FileActionModel must be staged to be
            # submitted through the API
            stage_method.append(f'{self.i2}data.indicator._staged = True  # noqa: SLF001')

        stage_method.extend(
            [
                f'{self.i2}self.model.{model_reference}.data.append(data)  # type: ignore',
                '',
                '',
            ]
        )

        return '\n'.join(stage_method)

    def _gen_code_object_remove_method(self) -> str:
        """Return the method code."""
        self.requirements['standard library'].append('import json')
        return '\n'.join(
            [
                f'{self.i1}def remove(self, params: dict | None = None):',
                f'{self.i2}"""Remove a nested object."""',
                f"{self.i2}method = 'PUT'",
                f'{self.i2}unique_id = self._calculate_unique_id()',
                '',
                f'{self.i2}# validate an id is available',
                f'{self.i2}self._validate_id(unique_id.get("value"), "")',
                '',
                f'{self.i2}body = json.dumps(',
                f'{self.i3}{{',
                f'{self.i4}self._nested_field_name: {{',
                f'{self.i5}"data": [{{unique_id.get("filter"): unique_id.get("value")}}],',
                f'{self.i5}"mode": "delete",',
                f'{self.i4}}}',
                f'{self.i3}}}',
                f'{self.i2})',
                '',
                f'{self.i2}# get the unique id value for id, xid, summary, etc ...',
                f'{self.i2}parent_api_endpoint = self._parent_data.get("api_endpoint")',
                f'{self.i2}parent_unique_id = self._parent_data.get("unique_id")',
                f'{self.i2}url = f"{{parent_api_endpoint}}/{{parent_unique_id}}"',
                '',
                f'{self.i2}# validate parent an id is available',
                f'{self.i2}self._validate_id(parent_unique_id, url)',
                '',
                f'{self.i2}self._request(',
                f'{self.i3}method=method,',
                f'{self.i3}url=url,',
                f'{self.i3}body=body,',
                f'{self.i3}headers={{"content-type": "application/json"}},',
                f'{self.i3}params=params,',
                f'{self.i2})',
                '',
                f'{self.i2}return self.request',
                '',
                '',
            ]
        )

    def _gen_code_object_stage_assignee(self) -> str:
        """Return the method code.

        def stage_assignee(
            self, type: str, data: dict | ObjectABC | AssigneeModel]
        ):
            '''Stage artifact on the object.'''
            if isinstance(data, ObjectABC):
                data = data.model
            elif type.lower() == 'user' and isinstance(data, dict):
                data = UserModel(**data)
            elif type.lower() == 'group' and isinstance(data, dict):
                data = UserGroupModel(**data)

            if not isinstance(data, UserModel | UserGroupModel):
                ex_msg = 'Invalid type passed in to stage_assignee'
                raise RuntimeError(ex_msg)  # noqa: TRY004
            data._staged = True  # noqa: SLF001
            self.model.assignee._staged = True  # noqa: SLF001
            self.model.assignee.type = type
            self.model.assignee.data = data
        """
        self.requirements['first-party'].extend(
            [
                'from tcex.api.tc.v3.security.user_groups.user_group_model import UserGroupModel',
                'from tcex.api.tc.v3.security.users.user_model import UserModel',
                'from tcex.api.tc.v3.security.assignee_model import AssigneeModel',
                (
                    'from tcex.api.tc.v3.security.assignee_user_group_model '
                    'import AssigneeUserGroupModel'
                ),
                'from tcex.api.tc.v3.security.assignee_user_model import AssigneeUserModel',
            ]
        )

        return '\n'.join(
            [
                f'{self.i1}',
                (
                    f'{self.i1}def stage_assignee(\n'
                    f'{self.i2}self,\n'
                    f'{self.i2}type: str,  # noqa: A002\n'
                    f'{self.i2}data: dict | ObjectABC | AssigneeModel '
                    f'| UserModel | UserGroupModel,\n'
                    f'{self.i1}):\n'
                ),
                f'{self.i2}"""Stage artifact on the object."""',
                f'{self.i2}if isinstance(data, ObjectABC):',
                f'{self.i3}data = data.model  # type: ignore',
                f'{self.i2}elif type.lower() == "user" and isinstance(data, dict):',
                f'{self.i3}data = AssigneeModel(type="User", data=data)  # type: ignore',
                f'{self.i2}elif type.lower() == "group" and isinstance(data, dict):',
                f'{self.i3}data = AssigneeModel(type="Group", data=data)  # type: ignore',
                '',
                f'{self.i2}if not isinstance(',
                f'{self.i3}data,',
                f'{self.i3}AssigneeModel | AssigneeUserModel | AssigneeUserGroupModel ',
                '| UserModel | UserGroupModel):',
                f'{self.i3}ex_msg = "Invalid type passed in to stage_assignee"',
                f'{self.i3}raise RuntimeError(ex_msg)  # noqa: TRY004',
                '',
                f'{self.i2}if isinstance(data, AssigneeModel):',
                f'{self.i3}self.model.assignee = data',
                f'{self.i2}elif isinstance(',
                f'{self.i3}data, AssigneeUserModel | AssigneeUserGroupModel '
                f'| UserModel | UserGroupModel',
                f'{self.i2}):',
                f'{self.i3}self.model.assignee.data = data  # type: ignore',
                # f'{self.i2}data._staged = True
                # f'{self.i2}self.model.assignee._staged = True
                # f'{self.i2}self.model.assignee.type = type',
                # pylance shows a warning on type here, but it in not handling inheritance properly.
                # f'{self.i2}self.model.assignee.data = data  # type: ignore',
                f'{self.i2}self.model.assignee.data._staged = True  # type: ignore  # noqa: SLF001',
                '',
                '',
            ]
        )

    def _gen_code_object_type_property_method(
        self, type_: str, model_type: str | None = None, custom_associations: bool = False
    ) -> str:
        """Return the method code.

        @property
        def artifacts(self):
            '''Yield Artifact from Artifacts'''
            from tcex.api.tc.v3.artifacts.model import Artifact, Artifacts

            yield from self._iterate_over_sublist(Artifacts)
        """
        type_ = self.util.snake_string(type_)
        model_type = self.util.snake_string(model_type or type_)

        # get model from map and update requirements
        model_import_data = self._module_import_data(type_)

        # Add Generator to imports:
        self.requirements['standard library'].append(
            {'module': 'collections.abc', 'imports': ['Generator']}
        )

        if self.type_ == type_:
            # set return type to new typing "Self" type and add import
            return_type = 'Self'
            self.requirements['standard library'].append(
                {'module': 'typing', 'imports': [return_type]}
            )
        else:
            # set typing to model class and add import
            return_type = f'"{model_import_data.get("object_class")}"'
            self.requirements['type-checking'].append(
                f'from {model_import_data.get("object_module")} '
                f'import {model_import_data.get("object_class")}'
            )

        _code = [
            f'{self.i1}@property',
            (f'{self.i1}def {model_type.plural()}(self) -> Generator[{return_type}, None, None]:'),
            (
                f'{self.i2}"""Yield {type_.singular().pascal_case()} '
                f'from {type_.plural().pascal_case()}."""'
            ),
        ]

        if self.type_ != type_:
            _code.extend(
                [
                    (
                        f'{self.i2}from {model_import_data.get("object_module")} '
                        f'import {model_import_data.get("object_collection_class")}'
                    ),
                    '',
                ]
            )

        # Custom logic to ensure that when iterating over the associated indicators or associated
        # groups then the item currently being iterated over is not included in the results.
        if custom_associations is True:
            _code.extend(
                [
                    (
                        f'{self.i2}yield from self._iterate_over_sublist'
                        f'({model_import_data.get("object_collection_class")}, '
                        'custom_associations=True)'
                        '  # type: ignore'
                    ),
                ]
            )
        elif (
            self.type_ in ['indicators', 'groups', 'artifacts', 'cases']
            and model_type == f'associated_{self.type_}'
        ):
            _code.extend(
                [
                    f'{self.i2}# Ensure the current item is not returned as a association',
                    (
                        f'{self.i2}for {type_.singular()} in self._iterate_over_sublist'
                        f'({model_import_data.get("object_collection_class")}):'
                        '  # type: ignore'
                    ),
                ]
            )
            if self.type_ == 'indicator':
                _code.extend(
                    [
                        (f'{self.i3}if {type_.singular()}.model.summary == self.model.summary:'),
                    ]
                )
            else:
                _code.extend(
                    [
                        (f'{self.i3}if {type_.singular()}.model.id == self.model.id:'),
                    ]
                )
            _code.extend(
                [
                    f'{self.i4}continue',
                    f'{self.i3}yield {type_.singular()}  # type: ignore',
                ]
            )
        else:
            _code.extend(
                [
                    (
                        f'{self.i2}yield from self._iterate_over_sublist'
                        f'({model_import_data.get("object_collection_class")})'
                        '  # type: ignore'
                    ),
                ]
            )
        _code.extend(
            [
                '',
                '',
            ]
        )
        return '\n'.join(_code)

    def gen_container_class(self) -> str:
        """Generate the Container Model

        class Artifacts(ObjectCollectionABC):
            '''Artifacts Collection.

            # Example of params input
            {
                'result_limit': 100,  # Limit the retrieved results.
                'result_start': 10,  # Starting count used for pagination.
                'fields': ['caseId', 'summary']  # Select additional return fields.
            }

            Arg:
                session (Session): Session object configured with TC API Auth.
                tql_filters (list): List of TQL filters.
                params (dict): Additional query params (see example above).
            '''
        """
        return '\n'.join(
            [
                '',
                f'class {self.type_.plural().pascal_case()}(ObjectCollectionABC):',
                f'{self.i1}"""{self.type_.plural().pascal_case()} Collection.',
                '',
                f'{self.i1}# Example of params input',
                f'{self.i1}{{',
                f'{self.i2}"result_limit": 100,  # Limit the retrieved results.',
                f'{self.i2}"result_start": 10,  # Starting count used for pagination.',
                f'{self.i2}"fields": ["caseId", "summary"]  # Select additional return fields.',
                f'{self.i1}}}',
                '',
                f'{self.i1}Args:',
                f'{self.i2}session (Session): Session object configured with TC API Auth.',
                f'{self.i2}tql_filters (list): List of TQL filters.',
                f'{self.i2}params (dict): Additional query params (see example above).',
                f'{self.i1}"""',
                '',
                '',
            ]
        )

    def gen_container_methods(self) -> str:
        """Return the container methods.

        def __init__ ...

        def __iter__ ...

        def filter ...

        """
        _code = ''
        # generate __init__ method
        _code += self._gen_code_container_init_method()

        # generate __iter__ method
        _code += self._gen_code_container_iter_method()

        # generate __next__ method
        # _code += self._gen_code_container_next_method()

        # generate api_endpoint property method
        _code += self._gen_code_api_endpoint_property()

        # generate filter property method
        _code += self._gen_code_container_filter_property()

        if self.type_.lower() == 'indicators':
            _code += self._gen_code_deleted_method()

        return _code

    def gen_doc_string(self) -> str:
        """Generate doc string."""
        # return (
        #     f'"""{self.type_.singular().pascal_case()} '
        #     f'/ {self.type_.plural().pascal_case()} Object"""\n'
        # )
        return '"""TcEx Framework Module"""\n'

    def gen_object_class(self) -> str:
        """Generate the Object Model

        class Artifact(ObjectABC):
            '''Case Management Artifact

            Arg:
                case_id (int, kwargs): The **case id** for the Artifact.
                case_xid (str, kwargs): The **case xid** for the Artifact.
                derived_link (bool, kwargs): Flag to specify if this artifact should be used for
                    potentially associated cases or not.
                ...
            '''
        """
        args = GenerateArgs(self.type_).gen_args()
        return '\n'.join(
            [
                '',
                f'class {self.type_.singular().pascal_case()}(ObjectABC):',
                f'''{self.i1}"""{self.type_.plural().pascal_case()} Object.''',
                '',
                f'{args}',
                f'{self.i1}"""',
                '',
                '',
            ]
        )

    def gen_object_methods(self) -> str:
        """Return the container methods.

        def __init__ ...

        def __iter__ ...

        def filter ...

        """
        _code = ''

        # generate __init__ method
        _code += self._gen_code_object_init_method()

        # generate api_endpoint property method
        _code += self._gen_code_api_endpoint_property()

        # generate model property
        _code += self._gen_code_object_model_property()

        # skip object that don't require as_entity method
        if self.type_ not in [
            'case_attributes',
            'victim_attributes',
            'group_attributes',
            'indicator_attributes',
            'security_labels',
            'tags',
            'attribute_types',
            'owner_roles',
            'owners',
            'system_roles',
            'user_groups',
            'users',
            'exclusion_lists',
        ]:
            # generate as_entity property method
            _code += self._gen_code_object_as_entity_property_method()

        # skip object that don't require remove method
        if self.type_ in [
            'groups',
            'indicators',
            'security_labels',
            'tags',
        ]:
            # generate remove property method
            _code += self._gen_code_object_remove_method()

        # generate group specific methods
        if self.type_.lower() == 'groups':
            _code += self._gen_code_group_methods()

        # get NON read-only properties of endpoint (OPTIONS: /v3/<object>)
        add_properties = [prop.name for prop in self.content_models if prop.read_only is False]

        # generate artifacts property method
        if 'artifacts' in add_properties:
            _code += self._gen_code_object_type_property_method('artifacts')

        # generate victim assets property method
        if 'assets' in add_properties:
            _code += self._gen_code_object_type_property_method('victim_assets')

        # generate associated_case property method
        if 'associatedArtifacts' in add_properties:
            _code += self._gen_code_object_type_property_method('artifacts', 'associated_artifacts')

        # generate associated_case property method
        if 'associatedCases' in add_properties:
            _code += self._gen_code_object_type_property_method('cases', 'associated_cases')

        # generate associated_group property method
        if 'associatedGroups' in add_properties:
            _code += self._gen_code_object_type_property_method('groups', 'associated_groups')

        # generate associated_indicator property method
        if 'associatedIndicators' in add_properties:
            _code += self._gen_code_object_type_property_method(
                'indicators', 'associated_indicators'
            )

        # generate custom_associations property method
        if 'customAssociations' in add_properties:
            _code += self._gen_code_object_type_property_method(
                'indicators', 'custom_associations', custom_associations=True
            )

        # generate associated_victim_asset property method
        if 'associatedVictimAssets' in add_properties:
            _code += self._gen_code_object_type_property_method(
                'victim_assets', 'associated_victim_assets'
            )

        # generate attributes property method
        if 'attributes' in add_properties:
            _code += self._gen_code_object_type_property_method('attributes')

        # generate cases property method
        if 'cases' in add_properties:
            _code += self._gen_code_object_type_property_method('cases')

        # generate notes property method
        if 'notes' in add_properties:
            _code += self._gen_code_object_type_property_method('notes')

        # generate security_labels property method
        if 'securityLabels' in add_properties:
            _code += self._gen_code_object_type_property_method('security_labels')

        # generate tags property method
        if 'tags' in add_properties:
            _code += self._gen_code_object_type_property_method('tags')

        # generate tasks property method
        if 'tasks' in add_properties:
            _code += self._gen_code_object_type_property_method('tasks')

        #
        # Stage Method
        #

        # generate stage_artifact method
        if 'artifacts' in add_properties:
            _code += self._gen_code_object_stage_type_method('artifacts')

        # [custom] generate stage assignee method
        if self.type_.lower() in ['cases', 'tasks'] and 'assignee' in add_properties:
            _code += self._gen_code_object_stage_assignee()

        # generate stage_asset method
        if 'assets' in add_properties:
            _code += self._gen_code_object_stage_type_method('victim_assets')

        # generate stage_associated_group method
        # ESUP-2532 - Associations are not Bi-Directional for IRs
        if 'associatedCases' in add_properties and self.type_ != 'intel_requirements':
            _code += self._gen_code_object_stage_type_method('cases', 'associated_cases')

        # ESUP-2532 - Associations are not Bi-Directional for IRs
        if 'associatedArtifacts' in add_properties and self.type_ != 'intel_requirements':
            _code += self._gen_code_object_stage_type_method('artifacts', 'associated_artifacts')

        # victims have associatedGroups but groups must be
        # associated to the asset not the victim object.
        if 'associatedGroups' in add_properties and self.type_.lower() not in ['victims']:
            _code += self._gen_code_object_stage_type_method('groups', 'associated_groups')

        # generate stage_associated_group method
        if 'associatedVictimAssets' in add_properties:
            _code += self._gen_code_object_stage_type_method(
                'victim_assets', 'associated_victim_assets'
            )

        # generate stage_associated_indicator method
        if 'associatedIndicators' in add_properties and self.type_ != 'indicators':
            _code += self._gen_code_object_stage_type_method('indicators', 'associated_indicators')

        # generate stage_attribute method
        if 'attributes' in add_properties:
            _code += self._gen_code_object_stage_type_method('attributes')

        # generate stage_case method
        if 'cases' in add_properties:
            _code += self._gen_code_object_stage_type_method('cases')

        # generate stage_file_actions method
        if 'fileActions' in add_properties:
            _code += self._gen_code_object_stage_type_method('file_actions')

        # generate stage_file_occurrences method
        if 'fileOccurrences' in add_properties:
            _code += self._gen_code_object_stage_type_method('file_occurrences')

        # generate stage_keyword_section method
        if 'keywordSections' in add_properties and self.type_ == 'intel_requirements':
            _code += self._gen_code_object_replace_type_method('keyword_sections')

        # generate stage_note method
        if 'notes' in add_properties:
            _code += self._gen_code_object_stage_type_method('notes')

        # generate stage_security_labels method
        if 'securityLabels' in add_properties:
            _code += self._gen_code_object_stage_type_method('security_labels')

        # generate stage_tag method
        if 'tags' in add_properties:
            _code += self._gen_code_object_stage_type_method('tags')

        # generate stage_task method
        if 'tasks' in add_properties:
            _code += self._gen_code_object_stage_type_method('tasks')

        # generate stage_task method
        if 'userAccess' in add_properties:
            _code += self._gen_code_object_stage_type_method('users', 'user_access')

        return _code

    def update_requirements(
        self,
        type_: SnakeString,
        filename: str,
        classes: list[str],
        from_: str = 'first-party',
    ):
        """Return the requirements code."""
        class_string = ', '.join(classes)
        self.requirements[from_].append(  # type: ignore
            f'from {self.tap(type_)}.{type_.plural()}.{filename} import {class_string}'
        )



"""TcEx Framework Module"""

from .app_spec_yml import AppSpecYml
from .install_json import InstallJson
from .job_json import JobJson
from .layout_json import LayoutJson
from .permutation import Permutation
from .tcex_json import TcexJson

__all__ = ['AppSpecYml', 'InstallJson', 'JobJson', 'LayoutJson', 'Permutation', 'TcexJson']



"""TcEx Framework Module"""

# standard library
from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    # first-party
    # from tcex.api.tc.v3.adversary_assets.adversary_asset import AdversaryAsset
    from tcex.api.tc.v3.artifact_types.artifact_type import ArtifactType
    from tcex.api.tc.v3.artifacts.artifact import Artifact
    from tcex.api.tc.v3.cases.case import Case
    from tcex.api.tc.v3.groups.group import Group
    from tcex.api.tc.v3.indicators.indicator import Indicator
    from tcex.api.tc.v3.notes.note import Note
    from tcex.api.tc.v3.security.owner_roles.owner_role import OwnerRole
    from tcex.api.tc.v3.security.owners.owner import Owner
    from tcex.api.tc.v3.security.system_roles.system_role import SystemRole
    from tcex.api.tc.v3.security.user_groups.user_group import UserGroup
    from tcex.api.tc.v3.security.users.user import User
    from tcex.api.tc.v3.tags.tag import Tag
    from tcex.api.tc.v3.tasks.task import Task
    from tcex.api.tc.v3.victim_assets.victim_asset import VictimAsset
    from tcex.api.tc.v3.victims.victim import Victim
    from tcex.api.tc.v3.workflow_events.workflow_event import WorkflowEvent
    from tcex.api.tc.v3.workflow_templates.workflow_template import WorkflowTemplate

    V3Type = (
        # AdversaryAsset
        ArtifactType
        | Artifact
        | Case
        | Group
        | Indicator
        | Note
        | OwnerRole
        | Owner
        | SystemRole
        | UserGroup
        | User
        | Tag
        | Task
        | VictimAsset
        | Victim
        | WorkflowEvent
        | WorkflowTemplate
    )



"""TcEx Framework Module"""

# standard library
import json
from collections.abc import Generator, Iterator
from typing import TYPE_CHECKING, Self

# third-party
from requests import Response

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.artifacts.artifact_model import ArtifactModel
from tcex.api.tc.v3.cases.case_model import CaseModel
from tcex.api.tc.v3.group_attributes.group_attribute_model import GroupAttributeModel
from tcex.api.tc.v3.groups.group_filter import GroupFilter
from tcex.api.tc.v3.groups.group_model import GroupModel, GroupsModel
from tcex.api.tc.v3.indicators.indicator_model import IndicatorModel
from tcex.api.tc.v3.object_abc import ObjectABC
from tcex.api.tc.v3.object_collection_abc import ObjectCollectionABC
from tcex.api.tc.v3.security_labels.security_label_model import SecurityLabelModel
from tcex.api.tc.v3.tags.tag_model import TagModel
from tcex.api.tc.v3.victim_assets.victim_asset_model import VictimAssetModel

if TYPE_CHECKING:  # pragma: no cover
    # first-party
    from tcex.api.tc.v3.artifacts.artifact import Artifact  # CIRCULAR-IMPORT
    from tcex.api.tc.v3.cases.case import Case  # CIRCULAR-IMPORT
    from tcex.api.tc.v3.group_attributes.group_attribute import GroupAttribute  # CIRCULAR-IMPORT
    from tcex.api.tc.v3.indicators.indicator import Indicator  # CIRCULAR-IMPORT
    from tcex.api.tc.v3.security_labels.security_label import SecurityLabel  # CIRCULAR-IMPORT
    from tcex.api.tc.v3.tags.tag import Tag  # CIRCULAR-IMPORT
    from tcex.api.tc.v3.victim_assets.victim_asset import VictimAsset  # CIRCULAR-IMPORT


class Group(ObjectABC):
    """Groups Object.

    Args:
        assignments (TaskAssignees, kwargs): A list of assignees and escalatees associated with this
            group (Task specific).
        associated_artifacts (Artifacts, kwargs): A list of Artifacts associated with this Group.
        associated_cases (Cases, kwargs): A list of Cases associated with this Group.
        associated_groups (Groups, kwargs): A list of groups associated with this group.
        associated_indicators (Indicators, kwargs): A list of indicators associated with this group.
        associated_victim_assets (VictimAssets, kwargs): A list of victim assets associated with
            this group.
        attributes (GroupAttributes, kwargs): A list of Attributes corresponding to the Group.
        body (str, kwargs): The email Body.
        due_date (str, kwargs): The date and time that the Task is due.
        escalation_date (str, kwargs): The escalation date and time.
        event_date (str, kwargs): The date and time that the incident or event was first created.
        event_type (str, kwargs): The identification of an event type.
        external_date_added (str, kwargs): The date and time that the item was first created
            externally.
        external_date_expires (str, kwargs): The date and time the item expires externally.
        external_last_modified (str, kwargs): The date and time the item was modified externally.
        file_name (str, kwargs): The document or signature file name.
        file_text (str, kwargs): The signature file text.
        file_type (str, kwargs): The signature file type.
        first_seen (str, kwargs): The date and time that the item was first seen.
        from_ (str, kwargs): The email From field.
        header (str, kwargs): The email Header field.
        last_seen (str, kwargs): The date and time that the item was last seen.
        malware (bool, kwargs): Is the document malware?
        name (str, kwargs): The name of the group.
        owner_id (int, kwargs): The id of the Organization, Community, or Source that the item
            belongs to.
        owner_name (str, kwargs): The name of the Organization, Community, or Source that the item
            belongs to.
        password (str, kwargs): The password associated with the document (Required if Malware is
            true).
        publish_date (str, kwargs): The date and time that the report was first created.
        reminder_date (str, kwargs): The reminder date and time.
        reviews (object, kwargs): A list of reviews corresponding to the Group.
        security_labels (SecurityLabels, kwargs): A list of Security Labels corresponding to the
            Intel item (NOTE: Setting this parameter will replace any existing tag(s) with
            the one(s) specified).
        status (str, kwargs): The status associated with this document, event, task, or incident
            (read only for task, document, and report).
        subject (str, kwargs): The email Subject section.
        tags (Tags, kwargs): A list of Tags corresponding to the item (NOTE: Setting this parameter
            will replace any existing tag(s) with the one(s) specified).
        type (str, kwargs): The **type** for the Group.
        xid (str, kwargs): The xid of the item.
    """

    def __init__(self, **kwargs):
        """Initialize instance properties."""
        super().__init__(kwargs.pop('session', None))

        # properties
        self._model: GroupModel = GroupModel(**kwargs)
        self._nested_field_name = 'associatedGroups'
        self._nested_filter = 'has_group'
        self.type_ = 'Group'

    @property
    def _api_endpoint(self) -> str:
        """Return the type specific API endpoint."""
        return ApiEndpoints.GROUPS.value

    @property
    def model(self) -> GroupModel:
        """Return the model data."""
        return self._model

    @model.setter
    def model(self, data: dict | GroupModel):
        """Create model using the provided data."""
        if isinstance(data, type(self.model)):
            # provided data is already a model, nothing required to change
            self._model = data
        elif isinstance(data, dict):
            # provided data is raw response, load the model
            self._model = type(self.model)(**data)
        else:
            ex_msg = f'Invalid data type: {type(data)} provided.'
            raise RuntimeError(ex_msg)  # noqa: TRY004

    @property
    def as_entity(self) -> dict:
        """Return the entity representation of the object."""
        type_ = self.model.type

        return {'type': type_, 'id': self.model.id, 'value': self.model.name}

    def remove(self, params: dict | None = None):
        """Remove a nested object."""
        method = 'PUT'
        unique_id = self._calculate_unique_id()

        # validate an id is available
        self._validate_id(unique_id.get('value'), '')

        body = json.dumps(
            {
                self._nested_field_name: {
                    'data': [{unique_id.get('filter'): unique_id.get('value')}],
                    'mode': 'delete',
                }
            }
        )

        # get the unique id value for id, xid, summary, etc ...
        parent_api_endpoint = self._parent_data.get('api_endpoint')
        parent_unique_id = self._parent_data.get('unique_id')
        url = f'{parent_api_endpoint}/{parent_unique_id}'

        # validate parent an id is available
        self._validate_id(parent_unique_id, url)

        self._request(
            method=method,
            url=url,
            body=body,
            headers={'content-type': 'application/json'},
            params=params,
        )

        return self.request

    def download(self, params: dict | None = None) -> bytes:
        """Return the document attachment for Document/Report Types."""
        self._request(
            method='GET',
            url=f'{self.url("GET")}/download',
            headers={'Accept': 'application/octet-stream'},
            params=params,
        )
        return self.request.content

    def pdf(self, params: dict | None = None) -> bytes:
        """Return the document attachment for Document/Report Types."""
        self._request(
            method='GET',
            body=None,
            url=f'{self.url("GET")}/pdf',
            headers={'Accept': 'application/octet-stream'},
            params=params,
        )

        return self.request.content

    def upload(self, content: bytes | str, params: dict | None = None) -> Response:
        """Return the document attachment for Document/Report Types."""
        self._request(
            method='POST',
            url=f'{self.url("GET")}/upload',
            body=content,
            headers={'content-type': 'application/octet-stream'},
            params=params,
        )
        return self.request

    @property
    def associated_artifacts(self) -> Generator['Artifact', None, None]:
        """Yield Artifact from Artifacts."""
        # first-party
        from tcex.api.tc.v3.artifacts.artifact import Artifacts

        yield from self._iterate_over_sublist(Artifacts)  # type: ignore

    @property
    def associated_cases(self) -> Generator['Case', None, None]:
        """Yield Case from Cases."""
        # first-party
        from tcex.api.tc.v3.cases.case import Cases

        yield from self._iterate_over_sublist(Cases)  # type: ignore

    @property
    def associated_groups(self) -> Generator[Self, None, None]:
        """Yield Group from Groups."""
        # Ensure the current item is not returned as a association
        for group in self._iterate_over_sublist(Groups):  # type: ignore
            if group.model.id == self.model.id:
                continue
            yield group  # type: ignore

    @property
    def associated_indicators(self) -> Generator['Indicator', None, None]:
        """Yield Indicator from Indicators."""
        # first-party
        from tcex.api.tc.v3.indicators.indicator import Indicators

        yield from self._iterate_over_sublist(Indicators)  # type: ignore

    @property
    def associated_victim_assets(self) -> Generator['VictimAsset', None, None]:
        """Yield VictimAsset from VictimAssets."""
        # first-party
        from tcex.api.tc.v3.victim_assets.victim_asset import VictimAssets

        yield from self._iterate_over_sublist(VictimAssets)  # type: ignore

    @property
    def attributes(self) -> Generator['GroupAttribute', None, None]:
        """Yield Attribute from Attributes."""
        # first-party
        from tcex.api.tc.v3.group_attributes.group_attribute import GroupAttributes

        yield from self._iterate_over_sublist(GroupAttributes)  # type: ignore

    @property
    def security_labels(self) -> Generator['SecurityLabel', None, None]:
        """Yield SecurityLabel from SecurityLabels."""
        # first-party
        from tcex.api.tc.v3.security_labels.security_label import SecurityLabels

        yield from self._iterate_over_sublist(SecurityLabels)  # type: ignore

    @property
    def tags(self) -> Generator['Tag', None, None]:
        """Yield Tag from Tags."""
        # first-party
        from tcex.api.tc.v3.tags.tag import Tags

        yield from self._iterate_over_sublist(Tags)  # type: ignore

    def stage_associated_case(self, data: dict | ObjectABC | CaseModel):
        """Stage case on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = CaseModel(**data)

        if not isinstance(data, CaseModel):
            ex_msg = 'Invalid type passed in to stage_associated_case'
            raise RuntimeError(ex_msg)  # noqa: TRY004
        data._staged = True  # noqa: SLF001
        self.model.associated_cases.data.append(data)  # type: ignore

    def stage_associated_artifact(self, data: dict | ObjectABC | ArtifactModel):
        """Stage artifact on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = ArtifactModel(**data)

        if not isinstance(data, ArtifactModel):
            ex_msg = 'Invalid type passed in to stage_associated_artifact'
            raise RuntimeError(ex_msg)  # noqa: TRY004
        data._staged = True  # noqa: SLF001
        self.model.associated_artifacts.data.append(data)  # type: ignore

    def stage_associated_group(self, data: dict | ObjectABC | GroupModel):
        """Stage group on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = GroupModel(**data)

        if not isinstance(data, GroupModel):
            ex_msg = 'Invalid type passed in to stage_associated_group'
            raise RuntimeError(ex_msg)  # noqa: TRY004
        data._staged = True  # noqa: SLF001
        self.model.associated_groups.data.append(data)  # type: ignore

    def stage_associated_victim_asset(self, data: dict | ObjectABC | VictimAssetModel):
        """Stage victim_asset on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = VictimAssetModel(**data)

        if not isinstance(data, VictimAssetModel):
            ex_msg = 'Invalid type passed in to stage_associated_victim_asset'
            raise RuntimeError(ex_msg)  # noqa: TRY004
        data._staged = True  # noqa: SLF001
        self.model.associated_victim_assets.data.append(data)  # type: ignore

    def stage_associated_indicator(self, data: dict | ObjectABC | IndicatorModel):
        """Stage indicator on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = IndicatorModel(**data)

        if not isinstance(data, IndicatorModel):
            ex_msg = 'Invalid type passed in to stage_associated_indicator'
            raise RuntimeError(ex_msg)  # noqa: TRY004
        data._staged = True  # noqa: SLF001
        self.model.associated_indicators.data.append(data)  # type: ignore

    def stage_attribute(self, data: dict | ObjectABC | GroupAttributeModel):
        """Stage attribute on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = GroupAttributeModel(**data)

        if not isinstance(data, GroupAttributeModel):
            ex_msg = 'Invalid type passed in to stage_attribute'
            raise RuntimeError(ex_msg)  # noqa: TRY004
        data._staged = True  # noqa: SLF001
        self.model.attributes.data.append(data)  # type: ignore

    def stage_security_label(self, data: dict | ObjectABC | SecurityLabelModel):
        """Stage security_label on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = SecurityLabelModel(**data)

        if not isinstance(data, SecurityLabelModel):
            ex_msg = 'Invalid type passed in to stage_security_label'
            raise RuntimeError(ex_msg)  # noqa: TRY004
        data._staged = True  # noqa: SLF001
        self.model.security_labels.data.append(data)  # type: ignore

    def stage_tag(self, data: dict | ObjectABC | TagModel):
        """Stage tag on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = TagModel(**data)

        if not isinstance(data, TagModel):
            ex_msg = 'Invalid type passed in to stage_tag'
            raise RuntimeError(ex_msg)  # noqa: TRY004
        data._staged = True  # noqa: SLF001
        self.model.tags.data.append(data)  # type: ignore


class Groups(ObjectCollectionABC):
    """Groups Collection.

    # Example of params input
    {
        "result_limit": 100,  # Limit the retrieved results.
        "result_start": 10,  # Starting count used for pagination.
        "fields": ["caseId", "summary"]  # Select additional return fields.
    }

    Args:
        session (Session): Session object configured with TC API Auth.
        tql_filters (list): List of TQL filters.
        params (dict): Additional query params (see example above).
    """

    def __init__(self, **kwargs):
        """Initialize instance properties."""
        super().__init__(
            kwargs.pop('session', None), kwargs.pop('tql_filter', None), kwargs.pop('params', None)
        )
        self._model = GroupsModel(**kwargs)
        self.type_ = 'groups'

    def __iter__(self) -> Iterator[Group]:
        """Return CM objects."""
        return self.iterate(base_class=Group)  # type: ignore

    @property
    def _api_endpoint(self) -> str:
        """Return the type specific API endpoint."""
        return ApiEndpoints.GROUPS.value

    @property
    def filter(self) -> GroupFilter:
        """Return the type specific filter object."""
        return GroupFilter(self.tql)



"""TcEx Framework Module"""

# standard library
from typing import ForwardRef

# third-party
from pydantic import BaseModel, validator
from pydantic.fields import ModelField  # TYPE-CHECKING

# first-party
from tcex.input.field_type.binary import Binary
from tcex.input.field_type.exception import InvalidEmptyValue
from tcex.input.field_type.sensitive import Sensitive
from tcex.input.field_type.string import String
from tcex.input.field_type.tc_entity import TCEntity

KeyValue = ForwardRef('KeyValue')  # type: ignore


class KeyValue(BaseModel):
    """Model for KeyValue Input."""

    key: str
    type: str | None
    value: (
        list[KeyValue]  # SELF-REFERENCE
        | KeyValue  # SELF-REFERENCE
        | list[TCEntity]
        | TCEntity
        | list[String]
        | String
        | list[Binary]
        | Binary
        | Sensitive
    )

    @validator('key')
    @classmethod
    def non_empty_string(cls, value: str, field: ModelField) -> str:
        """Validate that the value is a non-empty string."""
        if isinstance(value, str) and value.replace(' ', '') == '':
            raise InvalidEmptyValue(field_name=field.name)
        return value

    class Config:
        """Model Config"""

        validate_assignment = True


KeyValue.update_forward_refs()


def key_value(allow_none=False):
    """Return configured instance of KeyValue model."""
    key_value_model = KeyValue
    if allow_none is True:

        class _KeyValue(KeyValue):
            value: (
                list[KeyValue]
                | KeyValue
                | list[TCEntity]
                | TCEntity
                | list[String]
                | String
                | list[Binary]
                | Binary
                | Sensitive
                | None
            )

        key_value_model = _KeyValue
    return key_value_model



"""TcEx Framework Module"""  # noqa: A005

# standard library
import re
from collections.abc import Generator

# third-party
from pydantic.fields import ModelField  # TYPE-CHECKING

# first-party
from tcex.input.field_type.exception import (
    InvalidEmptyValue,
    InvalidLengthValue,
    InvalidPatternValue,
    InvalidType,
)


class String(str):
    """String Field Type"""

    allow_empty: bool = True
    max_length: int | None = None
    min_length: int | None = None
    regex: str | None = None
    strip: bool = False

    @classmethod
    def __get_validators__(cls) -> Generator:
        """Run validators / modifiers on input."""
        yield cls.validate_type
        yield cls.validate_strip
        yield cls.validate_allow_empty
        yield cls.validate_max_length
        yield cls.validate_min_length
        yield cls.validate_regex

    @classmethod
    def validate_allow_empty(cls, value: str, field: ModelField) -> str:
        """Raise exception if value is empty and allow_empty is False."""
        if cls.allow_empty is False and isinstance(value, str) and value == '':
            raise InvalidEmptyValue(field_name=field.name)

        return value

    @classmethod
    def validate_max_length(cls, value: str, field: ModelField) -> str:
        """Raise exception if value does not match pattern."""
        if cls.max_length is not None and len(value) > cls.max_length:
            raise InvalidLengthValue(
                field_name=field.name, constraint=cls.max_length, operation='max'
            )
        return value

    @classmethod
    def validate_min_length(cls, value: str, field: ModelField) -> str:
        """Raise exception if value does not match pattern."""
        if cls.min_length is not None and len(value) < cls.min_length:
            raise InvalidLengthValue(
                field_name=field.name, constraint=cls.min_length, operation='min'
            )
        return value

    @classmethod
    def validate_regex(cls, value: str, field: ModelField) -> str:
        """Raise exception if value does not match pattern."""
        if isinstance(cls.regex, str) and not re.compile(cls.regex).match(value):
            raise InvalidPatternValue(field_name=field.name, pattern=cls.regex)
        return value

    @classmethod
    def validate_strip(cls, value: str) -> str:
        """Raise exception if value is not a Binary type."""
        if cls.strip is True:
            value = value.strip()
        return value

    @classmethod
    def validate_type(cls, value: str, field: ModelField) -> str:
        """Raise exception if value is not a String type."""
        if not isinstance(value, str):
            raise InvalidType(
                field_name=field.name, expected_types='(str)', provided_type=type(value)
            )
        return value


def string(
    allow_empty: bool = True,
    max_length: int | None = None,
    min_length: int | None = None,
    regex: str | None = None,
    strip: bool = False,
) -> type[String]:
    """Return configured instance of String."""
    namespace = {
        'allow_empty': allow_empty,
        'max_length': max_length,
        'min_length': min_length,
        'regex': regex,
        'strip': strip,
    }
    return type('ConstrainedString', (String,), namespace)



"""TcEx Framework Module"""

# standard library
from typing import ClassVar

# third-party
from pydantic import create_model, validator
from pydantic.fields import ModelField  # TYPE-CHECKING

# first-party
from tcex.api.tc.util.threat_intel_util import ThreatIntelUtil
from tcex.input.field_type.exception import InvalidEmptyValue, InvalidEntityType
from tcex.input.field_type.tc_entity import TCEntity
from tcex.registry import registry


class GroupEntity(TCEntity):
    """Group Entity Field (Model) Type"""

    group_types: ClassVar[list[str]] = []

    @validator('type', allow_reuse=True)
    @classmethod
    def is_empty(cls, value: str, field: ModelField) -> str:
        """Validate that the value is a non-empty string."""
        if isinstance(value, str) and value.replace(' ', '') == '':
            raise InvalidEmptyValue(field_name=field.name)
        return value

    @validator('type', allow_reuse=True)
    @classmethod
    def is_type(cls, value: str, field: ModelField) -> str:
        """Validate that the value is a non-empty string.

        Without the always and pre args, None values will validated before this validator is called.
        """
        ti_utils = ThreatIntelUtil(session_tc=registry.session_tc)
        group_types = cls.group_types or ti_utils.group_types
        if value.lower() not in [i.lower() for i in group_types]:
            raise InvalidEntityType(
                field_name=field.name, entity_type=str(group_types), value=value
            )
        return value


def group_entity(
    group_types: list[str], model_name: str = 'CustomGroupEntity'
) -> type[GroupEntity]:
    """Dynamically create a Case Management Entity model."""
    return create_model(
        model_name,
        group_types=(ClassVar[list[str]], group_types),
        __base__=GroupEntity,
    )


AdversaryEntity = group_entity(group_types=['Adversary'])
CampaignEntity = group_entity(group_types=['Campaign'])
DocumentEntity = group_entity(group_types=['Document'])
EmailEntity = group_entity(group_types=['Email'])
EventEntity = group_entity(group_types=['Event'])
IncidentEntity = group_entity(group_types=['Incident'])
IntrusionSetEntity = group_entity(group_types=['Intrusion Set'])
SignatureEntity = group_entity(group_types=['Signature'])
ReportEntity = group_entity(group_types=['Report'])
ThreatEntity = group_entity(group_types=['Threat'])
TaskEntity = group_entity(group_types=['Task'])



"""TcEx Framework Module"""

# standard library
from datetime import datetime

# first-party
from tcex.api.tc.ti_transform.model import GroupTransformModel, IndicatorTransformModel
from tcex.api.tc.ti_transform.transform_abc import (
    NoValidTransformException,
    TransformABC,
    TransformException,
    TransformsABC,
)


class TiTransforms(TransformsABC):
    """Mappings"""

    def process(self):
        """Process the mapping."""
        self.transformed_collection: list[TiTransform] = []
        for ti_dict in self.ti_dicts:
            self.transformed_collection.append(TiTransform(ti_dict, self.transforms))

    @property
    def batch(self) -> dict:
        """Return the data in batch format."""
        self.process()

        batch = {
            'group': [],
            'indicator': [],
        }
        self.log.trace(f'feature=ti-transform-batch, ti-count={len(self.transformed_collection)}')
        for index, t in enumerate(self.transformed_collection):
            if index and index % 1_000 == 0:
                self.log.trace(f'feature=ti-transform-batch, items={index}')

            # batch must be called so that the transform type is selected
            try:
                data = t.batch
            except NoValidTransformException:
                self.log.exception('feature=ti-transforms, event=runtime-error')
                continue
            except TransformException as e:
                self.log.warning(
                    f'feature=ti-transforms, event=transform-error, field="{e.field}", '
                    f'cause="{e.cause}", context="{e.context}"'
                )
                if self.raise_exceptions:
                    raise
                continue
            except Exception:
                self.log.exception('feature=ti-transforms, event=transform-error')
                if self.raise_exceptions:
                    raise
                continue

            # now that batch is called we can identify the ti type
            if isinstance(t.transform, GroupTransformModel):
                batch['group'].append(data)
            elif isinstance(t.transform, IndicatorTransformModel):
                batch['indicator'].append(data)

            # append adhoc groups and indicators
            batch['group'].extend(t.adhoc_groups)
            batch['indicator'].extend(t.adhoc_indicators)
        return batch


class TiTransform(TransformABC):
    """Threat Intelligence Transform Module"""

    def add_associated_group(self, group_xid: str):
        """Add an associated group.

        {
            'associatedGroups': [
                {
                    'groupXid': 'dd78f2b94ac61d3e5a55c1223a7635db00cd0aaa8aba26c5306e36dd6c1662ee'}
        """
        # process type specific data
        if isinstance(self.transform, GroupTransformModel):
            self.transformed_item.setdefault('associatedGroupXid', []).append(group_xid)
        elif isinstance(self.transform, IndicatorTransformModel):
            associated_group = {'groupXid': group_xid}
            self.transformed_item.setdefault('associatedGroups', []).append(associated_group)

    def add_attribute(
        self,
        type_: str,
        value: str,
        displayed: bool = False,
        pinned: bool = False,
        source: str | None = None,
    ):
        """Add an attribute to the transformed item."""
        if type_ is not None and value is not None:
            attribute_data: dict[str, bool | str] = {
                'type': type_,
                'value': value,
            }

            # displayed is a special case, it only needs to be added if True
            if displayed is True:
                attribute_data['displayed'] = displayed

            # pinned is a special case, it only needs to be added if True
            if pinned is True:
                attribute_data['pinned'] = displayed

            # source is a special case, it only needs to be added if not None
            if source is not None:
                attribute_data['source'] = source

            self.transformed_item.setdefault('attribute', []).append(attribute_data)

    def add_file_occurrence(
        self,
        file_name: str | None = None,
        path: str | None = None,
        date: datetime | None = None,
    ):
        """Abstract method"""
        self.transformed_item.setdefault('fileOccurrence', []).append(
            {
                k: v
                for k, v in {
                    'fileName': file_name,
                    'path': path,
                    'date': date,
                }.items()
                if v
            }
        )

    def add_confidence(self, confidence: int | str | None):
        """Add a rating to the transformed item."""
        if confidence is not None:
            self.transformed_item['confidence'] = int(confidence)

    def add_group(self, group_data: dict):
        """Add a group to the transforms.

        Group data must match the format of the endpoint being used, (e.g., batch format
        for batch endpoints, v3 format for v3 endpoints).
        """
        self.adhoc_groups.append(group_data)

    def add_indicator(self, indicator_data: dict):
        """Add a indicator to the transforms.

        Indicator data must match the format of the endpoint being used, (e.g., batch format
        for batch endpoints, v3 format for v3 endpoints).
        """
        self.adhoc_indicators.append(indicator_data)

    def add_metadata(self, key: str, value: str):
        """Add name to the transformed item."""
        if all([key, value]):
            self.transformed_item[key] = value

    def add_name(self, name: str | None):
        """Add name to the transformed item."""
        if name is not None:
            self.transformed_item['name'] = name

    def add_rating(self, rating: float | str | None):
        """Add a rating to the transformed item."""
        if rating is not None:
            self.transformed_item['rating'] = float(rating)

    def add_security_label(
        self, name: str, color: str | None = None, description: str | None = None
    ):
        """Add a tag to the transformed item."""
        if name is not None:
            label_data = {'name': name}

            if color is not None:
                label_data['color'] = color

            if description is not None:
                label_data['description'] = description

            self.transformed_item.setdefault('securityLabel', []).append(label_data)

    def add_summary(self, value: str | None):
        """Add value1 to the transformed item."""
        if value is not None:
            self.transformed_item['summary'] = value

    def add_tag(self, name: str):
        """Add a tag to the transformed item."""
        if name is not None:
            self.transformed_item.setdefault('tag', []).append({'name': name})

    @property
    def batch(self) -> dict:
        """Return the data in batch format."""
        self._process()
        return dict(sorted(self.transformed_item.items()))



"""TcEx Framework Module"""

# standard library
from enum import Enum

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.filter_abc import FilterABC
from tcex.api.tc.v3.tql.tql_type import TqlType


class ArtifactTypeFilter(FilterABC):
    """Filter Object for ArtifactTypes"""

    @property
    def _api_endpoint(self) -> str:
        """Return the API endpoint."""
        return ApiEndpoints.ARTIFACT_TYPES.value

    def active(self, operator: Enum, active: bool):
        """Filter Active based on **active** keyword.

        Args:
            operator: The operator enum for the filter.
            active: The active status of the artifact type.
        """
        self._tql.add_filter('active', operator, active, TqlType.BOOLEAN)

    def data_type(self, operator: Enum, data_type: list | str):
        """Filter Data Type based on **dataType** keyword.

        Args:
            operator: The operator enum for the filter.
            data_type: The data type of the artifact type.
        """
        if isinstance(data_type, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('dataType', operator, data_type, TqlType.STRING)

    def description(self, operator: Enum, description: list | str):
        """Filter Description based on **description** keyword.

        Args:
            operator: The operator enum for the filter.
            description: The description of the artifact type.
        """
        if isinstance(description, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('description', operator, description, TqlType.STRING)

    def id(self, operator: Enum, id: int | list):  # noqa: A002
        """Filter ID based on **id** keyword.

        Args:
            operator: The operator enum for the filter.
            id: The ID of the artifact type.
        """
        if isinstance(id, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('id', operator, id, TqlType.INTEGER)

    def intel_type(self, operator: Enum, intel_type: list | str):
        """Filter Intel Type based on **intelType** keyword.

        Args:
            operator: The operator enum for the filter.
            intel_type: The intel type of the artifact type.
        """
        if isinstance(intel_type, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('intelType', operator, intel_type, TqlType.STRING)

    def managed(self, operator: Enum, managed: bool):
        """Filter Managed based on **managed** keyword.

        Args:
            operator: The operator enum for the filter.
            managed: The managed status of the artifact type.
        """
        self._tql.add_filter('managed', operator, managed, TqlType.BOOLEAN)

    def name(self, operator: Enum, name: list | str):
        """Filter Name based on **name** keyword.

        Args:
            operator: The operator enum for the filter.
            name: The name of the artifact type.
        """
        if isinstance(name, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('name', operator, name, TqlType.STRING)



"""TcEx Framework Module"""

# standard library
import logging
import re
from urllib.parse import quote

# third-party
import jmespath
from requests import Session

# first-party
from tcex.exit.error_code import handle_error
from tcex.logger.trace_logger import TraceLogger
from tcex.pleb.cached_property import cached_property

# get tcex logger
_logger: TraceLogger = logging.getLogger(__name__.split('.', maxsplit=1)[0])  # type: ignore


class ThreatIntelUtil:
    """Threat Intelligence Common Methods"""

    INDICATOR = 'Indicator'
    GROUP = 'Group'

    def __init__(self, session_tc: Session):
        """Initialize instance properties."""
        self.session_tc = session_tc

        # properties
        self.log = _logger
        self._ti = None

    @property
    def resolvable_variables(self) -> dict:
        """Return a dict of all the supported resolvable variables.

        Each entry in the dict has the variables corresponding url and jmespath for their values
        if available. Some variables do not have this data since no api endpoint is supported.
        """

        return {
            '${API_USERS}': {'url': '/v3/security/users', 'jmspath': 'data[*].userName'},
            '${ARTIFACT_TYPES}': {'url': '/v3/artifactTypes', 'jmspath': 'data[*].name'},
            '${ATTRIBUTES}': {'url': '/v3/attributeTypes', 'jmspath': 'data[*].name'},
            '${GROUP_TYPES}': None,  # Special Case since there is no api endpoint
            '${INDICATOR_TYPES}': {
                'url': '/v2/types/indicatorTypes',
                'jmspath': 'data.indicatorType[*].name',
            },
            '${OWNERS}': {'url': '/v3/security/owners', 'jmspath': 'data[*].name'},
            '${USER_GROUPS}': {'url': '/v3/security/userGroups', 'jmspath': 'data[*].name'},
            '${USERS}': {'url': '/v3/security/users', 'jmspath': 'data[*].userName'},
            '${WORKFLOW_TEMPLATES}': {
                'url': '/v3/workflowTemplates?tql=active = true',
                'jmspath': 'data[*].name',
            },
        }

    def _association_types(self):
        """Retrieve Custom Indicator Associations types from the ThreatConnect API."""

    @staticmethod
    def expand_indicators(indicator: str) -> list[str]:
        """Process indicators expanding file hashes/custom indicators into multiple entries.

        Args:
            indicator: A " : " delimited string.

        Returns:
            A list of indicators split on " : ".
        """
        if indicator.count(' : ') > 0:
            # handle all multi-valued indicators types (file hashes and custom indicators)
            indicator_list = []

            iregx_pattern = (
                # group 1 - lazy capture everything to first <space>:<space> or end of line
                r'^(.*?(?=\s\:\s|$))?'
                r'(?:\s\:\s)?'  # remove <space>:<space>
                # group 2 - look behind for <space>:<space>, lazy capture everything
                # to look ahead (optional <space>):<space> or end of line
                r'((?<=\s\:\s).*?(?=(?:\s)?\:\s|$))?'
                r'(?:(?:\s)?\:\s)?'  # remove (optional <space>):<space>
                # group 3 - look behind for <space>:<space>, lazy capture everything
                # to look ahead end of line
                r'((?<=\s\:\s).*?(?=$))?$'
            )
            iregx = re.compile(iregx_pattern)

            indicators = iregx.search(indicator)
            if indicators is not None:
                indicator_list = list(indicators.groups())
        else:
            # handle all single valued indicator types (address, host, etc)
            indicator_list = [indicator]

        return indicator_list

    @property
    def group_types(self) -> list[str]:
        """Return all defined ThreatConnect Group types.

        Returns:
            A list of ThreatConnect Group types.
        """
        return list(self.group_types_data.keys())

    @property
    def group_types_data(self) -> dict[str, dict]:
        """Return supported ThreatConnect Group types."""
        return {
            'Adversary': {'apiBranch': 'adversaries', 'apiEntity': 'adversary'},
            'Attack Pattern': {'apiBranch': 'attackPatterns', 'apiEntity': 'attackPattern'},
            'Campaign': {'apiBranch': 'campaigns', 'apiEntity': 'campaign'},
            'Course of Action': {'apiBranch': 'coursesOfAction', 'apiEntity': 'courseOfAction'},
            'Document': {'apiBranch': 'documents', 'apiEntity': 'document'},
            'Email': {'apiBranch': 'emails', 'apiEntity': 'email'},
            'Event': {'apiBranch': 'events', 'apiEntity': 'event'},
            'Incident': {'apiBranch': 'incidents', 'apiEntity': 'incident'},
            'Intrusion Set': {'apiBranch': 'intrusionSets', 'apiEntity': 'intrusionSet'},
            'Malware': {'apiBranch': 'malware', 'apiEntity': 'malware'},
            'Report': {'apiBranch': 'reports', 'apiEntity': 'report'},
            'Signature': {'apiBranch': 'signatures', 'apiEntity': 'signature'},
            'Tactic': {'apiBranch': 'tactics', 'apiEntity': 'tactic'},
            'Task': {'apiBranch': 'tasks', 'apiEntity': 'task'},
            'Threat': {'apiBranch': 'threats', 'apiEntity': 'threat'},
            'Tool': {'apiBranch': 'tools', 'apiEntity': 'tool'},
            'Vulnerability': {'apiBranch': 'vulnerabilities', 'apiEntity': 'vulnerability'},
        }

    def get_type_from_api_entity(self, api_entity: dict) -> str | None:
        """Return the object type as a string given a api entity.

        Args:
            api_entity: A TCEntity object.

        Returns:
            str, None: The type value or None.

        """
        merged = self.group_types_data.copy()
        merged.update(self.indicator_types_data)
        for key, value in merged.items():
            if value.get('apiEntity') == api_entity:
                return key
        return None

    @cached_property
    def indicator_associations_types_data(self) -> dict[str, dict]:
        """Return ThreatConnect associations type data.

        Retrieve the data from the API if it hasn't already been retrieved.

        Returns:
            (dict): A dictionary of ThreatConnect associations types.
        """
        _association_types = {}

        # Dynamically create custom indicator class
        r = self.session_tc.get('/v2/types/associationTypes')

        # check for bad status code and response that is not JSON
        if not r.ok or 'application/json' not in r.headers.get('content-type', ''):
            self.log.warning(
                'feature=threat-intel-common, event=association-types-download, status=failure'
            )
            return _association_types

        # validate successful API results
        data: dict = r.json()
        if data.get('status') != 'Success':
            self.log.warning(
                'feature=threat-intel-common, event=association-types-download, status=failure'
            )
            return _association_types

        # TODO: [low] make an Model for this data and return model?
        try:
            # Association Type Name is not a unique value at this time, but should be.
            association_types = {}
            for association in data.get('data', {}).get('associationType', []):
                association_types[association.get('name')] = association
        except Exception as e:
            handle_error(code=200, message_values=[e])
        return _association_types

    @cached_property
    def indicator_types(self) -> list[str]:
        """Return ThreatConnect Indicator types.

        Retrieve the data from the API if it hasn't already been retrieved.

        Returns:
            (list): A list of ThreatConnect Indicator types.
        """
        return list(self.indicator_types_data.keys())

    def resolve_variables(self, inputs: list[str]) -> list[str]:
        """Resolve all of the provided inputs if appropriate.

        Args:
            inputs: A list of strings to resolve if string matches a entry in the
            "resolved_variables" dict.
        """
        resolved_inputs = []
        for input_ in inputs:
            # handle null inputs
            if not input_:
                resolved_inputs.append(None)
                continue

            # clean up input
            input__ = input_.strip()

            # handle unknown input types
            if input__ not in self.resolvable_variables:
                resolved_inputs.append(input__)
                continue

            # special handling of group types (no API request required)
            if input__ == '${GROUP_TYPES}':
                for type_ in self.group_types:
                    resolved_inputs.append(type_)
                continue

            # get variable settings
            resolvable_variable_details = self.resolvable_variables[input__]

            # make API call to retrieve variable data
            r = self.session_tc.get(
                resolvable_variable_details.get('url'), params={'resultLimit': 10_000}
            )

            if not r.ok:
                ex_msg = f'Could not retrieve {input__} from ThreatConnect API.'
                raise RuntimeError(ex_msg)

            json_ = r.json()
            for item in jmespath.search(resolvable_variable_details.get('jmspath'), json_):
                resolved_inputs.append(str(item))

        return resolved_inputs

    @cached_property
    def indicator_types_data(self) -> dict[str, dict]:
        """Return ThreatConnect indicator types data.

        Retrieve the data from the API if it hasn't already been retrieved.

        Returns:
            (dict): A dictionary of ThreatConnect Indicator data.
        """
        # retrieve data from API
        r = self.session_tc.get('/v2/types/indicatorTypes')

        # TODO: [low] use handle error instead
        if not r.ok:
            ex_msg = 'Could not retrieve indicator types from ThreatConnect API.'
            raise RuntimeError(ex_msg)

        _indicator_types = {}
        for itd in r.json().get('data', {}).get('indicatorType'):
            _indicator_types[itd.get('name')] = itd
        return _indicator_types

    @staticmethod
    def safe_indicator(indicator: str) -> str:
        """Format indicator value for safe HTTP request.

        Args:
            indicator: Indicator to URL Encode

        Returns:
            (str): The urlencoded string
        """
        if indicator is not None:
            indicator = quote(indicator, safe='~')
        return indicator

    @staticmethod
    def safe_rt(resource_type: str, lower: bool = False) -> str:
        """Format the Resource Type.

        Takes Custom Indicator types with a space character and return a *safe* string.

        (e.g. *User Agent* is converted to User_Agent or user_agent.)

        Args:
           resource_type: The resource type to format.
           lower: Return type in all lower case

        Returns:
            (str): The formatted resource type.
        """
        if resource_type is not None:
            resource_type = resource_type.replace(' ', '_')
            if lower:
                resource_type = resource_type.lower()
        return resource_type

    @staticmethod
    def safe_group_name(group_name: str, group_max_length: int = 100, ellipsis: bool = True) -> str:
        """Truncate group name to match limit breaking on space and optionally add an ellipsis.

        .. note:: Currently the ThreatConnect group name limit is 100 characters.

        Args:
           group_name: The raw group name to be truncated.
           group_max_length: The max length of the group name.
           ellipsis: If true the truncated name will have '...' appended.

        Returns:
            (str): The truncated group name with optional ellipsis.
        """
        ellipsis_value = ''
        if ellipsis:
            ellipsis_value = ' ...'

        if group_name is not None and len(group_name) > group_max_length:
            # split name by spaces and reset group_name
            group_name_array = group_name.split(' ')
            group_name = ''
            for word in group_name_array:
                word_ = f'{word}'
                if (len(group_name) + len(word_) + len(ellipsis_value)) >= group_max_length:
                    group_name = f'{group_name}{ellipsis_value}'
                    group_name = group_name.lstrip(' ')
                    break
                group_name += f' {word_}'
        return group_name

    @staticmethod
    def safe_tag(tag: str) -> str:
        """Encode and truncate tag to match limit (128 characters) of ThreatConnect API.

        Args:
           tag: The tag to be truncated

        Returns:
            (str): The truncated and quoted tag.
        """
        if tag is not None:
            tag = quote(tag[:128], safe='~')
        return tag

    @staticmethod
    def safe_url(url: str) -> str:
        """Encode value for safe HTTP request.

        Args:
            url (str): The string to URL Encode.

        Returns:
            (str): The urlencoded string.
        """
        if url is not None:
            url = quote(url, safe='~')
        return url

    @property
    def victim_asset_types(self) -> list:
        """Return all defined ThreatConnect Asset types.

        Returns:
            (list): A list of ThreatConnect Asset types.
        """
        return [
            'EmailAddress',
            'SocialNetwork',
            'NetworkAccount',
            'WebSite',
            'Phone',
        ]

    def validate_intel_types(self, types_list: list[str], restrict_to: str | None = None):
        """Validate that Types contained in types_list are valid Intel Types.

        :param types_list: list of types to validate. An exception is raised if a member of this
        list is not a valid intel type.

        :param restrict_to: If None, types_list will be validated to contain valid Indicator
        or Group types. If not None, this value must be set to self.INDICATOR or self.GROUP,
        and types_list will be validated to contain only Indicators or Groups depending on this
        value.
        """

        if restrict_to is not None and restrict_to not in [self.INDICATOR, self.GROUP]:
            ex_msg = f'restrict_to must be {self.INDICATOR} or {self.GROUP}'
            raise ValueError(ex_msg)

        if not isinstance(types_list, list):
            ex_msg = 'types_list must be a a list.'
            raise TypeError(ex_msg)

        if restrict_to:
            valid_types = self.group_types if restrict_to == self.GROUP else self.indicator_types
        else:
            valid_types = self.indicator_types + self.group_types

        if any(_type not in valid_types for _type in types_list):
            ex_msg = f'Type list "{types_list}" contains invalid Intel Type.'
            raise ValueError(ex_msg)



"""TcEx Framework Module"""

# standard library
import sys
from typing import Literal, NoReturn

# third-party
from rich import print as print_
from rich.columns import Columns
from rich.panel import Panel
from rich.rule import Rule
from rich.text import Text

AlignMethod = Literal['left', 'center', 'right']


class RenderPanel:
    """Render Panel"""

    accent = 'dark_orange'
    title_align: AlignMethod = 'left'

    @classmethod
    def column(
        cls,
        column_data: list[str],
        title: str,
        subtitle: str | None = None,
        equal: bool = False,
        expand: bool = True,
    ):
        """Render Column with index in a panel."""
        columns = Columns(column_data, equal=equal, expand=expand)
        print_(
            Panel(
                columns,
                title=title,
                title_align=cls.title_align,
                subtitle=subtitle,
            )
        )

    @classmethod
    def column_index(
        cls,
        column_data: list[str],
        title: str,
        subtitle: str | None = None,
        equal: bool = False,
        expand: bool = True,
    ):
        """Render Column with index in a panel."""
        column_data = [f'[{cls.accent}]{i}.[/{cls.accent}] {o}' for i, o in enumerate(column_data)]
        columns = Columns(column_data, equal=equal, expand=expand)
        print_(
            Panel(
                columns,
                title=title,
                title_align=cls.title_align,
                subtitle=subtitle,
            )
        )

    @classmethod
    def error(cls, message: str):
        """Render error panel."""
        print_(
            Panel(
                Text(
                    f'{message}',
                    style='bold red',
                ),
                expand=True,
                title='Error',
                title_align=cls.title_align,
            )
        )

    @classmethod
    def failure(cls, message: str) -> NoReturn:
        """Render failure panel."""
        print_(
            Panel(
                Text(
                    f'{message}',
                    style='bold red',
                ),
                expand=True,
                title='Failure',
                title_align=cls.title_align,
            )
        )
        sys.exit(1)

    @staticmethod
    def info(message: str, title: str = 'Info', title_align: AlignMethod = 'left'):
        """Render info panel."""
        print_(
            Panel(
                f'{message}',
                expand=True,
                title=title,
                title_align=title_align,
            )
        )

    @staticmethod
    def invalid_value(message: str, title: str, title_align: AlignMethod = 'left'):
        """Render invalid values panel."""
        print_(
            Panel(
                Text(
                    f'{message}',
                    style='bold yellow',
                ),
                expand=True,
                title=title,
                title_align=title_align,
            )
        )

    @classmethod
    def list(cls, title: str, items: list[str], style: str = ''):
        """Render list panel."""
        items = [f'[white]•[/white] {i}' for i in items]
        item_list = '\n'.join(items)

        # render error panel
        if item_list:
            print_(
                Panel(
                    item_list,
                    border_style='',
                    style=style,
                    title=title,
                    title_align=cls.title_align,
                )
            )

    @staticmethod
    def rule(title: str, align: AlignMethod = 'center', style: str = ''):
        """Render a horizontal rule."""
        print_(
            Panel(
                Rule(
                    title=title,
                    align=align,
                    style=style,
                )
            )
        )

    @classmethod
    def success(cls, message: str):
        """Render success panel."""
        print_(
            Panel(
                Text(
                    f'{message}',
                    style='bold green',
                ),
                expand=True,
                title='Success',
                title_align=cls.title_align,
            )
        )

    @classmethod
    def warning(cls, message: str):
        """Render warning panel."""
        print_(
            Panel(
                Text(
                    f'{message}',
                    style='bold yellow',
                ),
                expand=True,
                title='Warning',
                title_align=cls.title_align,
            )
        )



"""TcEx Framework Module"""

# standard library
from urllib.parse import quote

from ..input.field_type.sensitive import Sensitive  # type: ignore


def proxies(
    proxy_host: str | None,
    proxy_port: int | None,
    proxy_user: str | None,
    proxy_pass: Sensitive | str | None,
) -> dict:
    """Format the proxy configuration for Python Requests module.

    Generates a dictionary for use with the Python Requests module format
    when proxy is required for remote connections.

    **Example Response**
    ::

        {
            'http': 'http://user:pass@10.10.1.10:3128/',
            'https': 'http://user:pass@10.10.1.10:3128/',
        }
    """
    _proxies = {}
    if proxy_host is not None and proxy_port is not None:
        proxy_auth = ''
        if proxy_user is not None and proxy_pass is not None:
            proxy_user = quote(proxy_user, safe='~')
            if isinstance(proxy_pass, Sensitive):
                proxy_pass_ = quote(proxy_pass.value, safe='~')  # type: ignore
            else:
                proxy_pass_ = quote(proxy_pass, safe='~')

            # proxy url with auth
            proxy_auth = f'{proxy_user}:{proxy_pass_}@'

        proxy_url = f'{proxy_auth}{proxy_host}:{proxy_port}'
        _proxies = {'http': f'http://{proxy_url}', 'https': f'http://{proxy_url}'}
    return _proxies



"""TcEx Framework Module"""

# standard library
import tempfile
import uuid
from pathlib import Path

# third-party
from pydantic import BaseModel, Field


class PathModel(BaseModel):
    """Path Model

    Supported for the following runtimeLevel:
    * ApiService
    * Playbook
    * Organization
    * WebhookTriggerService
    * TriggerService
    """

    #
    # ThreatConnect Provided Inputs
    #

    tc_in_path: Path = Field(
        Path(tempfile.gettempdir()),
        description='The path to the Apps "in" directory.',
        inclusion_reason='runtimeLevel',
    )
    tc_log_path: Path = Field(
        Path(tempfile.gettempdir()),
        description='The path to the Apps "log" directory.',
        inclusion_reason='runtimeLevel',
    )
    tc_out_path: Path = Field(
        Path(tempfile.gettempdir()),
        description='The path to the Apps "out" directory.',
        inclusion_reason='runtimeLevel',
    )
    tc_temp_path: Path = Field(
        Path(tempfile.gettempdir()),
        description='The path to the Apps "tmp" directory.',
        inclusion_reason='runtimeLevel',
    )

    @property
    def tc_session_id(self) -> str:
        """Return the current session id."""
        if self.tc_out_path.parent.name:
            return self.tc_in_path.parent.name

        # handle running locally
        session_id_filename = self.tc_out_path / 'tc_session_id'
        if not session_id_filename.exists():
            session_id_filename.write_text(str(uuid.uuid4()))
        return session_id_filename.read_text().strip()
